package de.zappbrannigan.ai;

import java.awt.Point;
import java.util.ArrayList;

import de.zappbrannigan.communication.FrameInfo;
import de.zappbrannigan.objects.*;

/**
 * Collection of helper methods used in calculations
 */
public final class Helper implements GameData {

	/**
	 * Calculates the intersection point of two straights
	 * 
	 * @param c1 Base point of line 1 (Ship)
	 * @param v1 Direction vector of line 1 (Ship)
	 * @param c2 Base point of line 2 (Target)
	 * @param v2 Direction vector of line 2 (Target)
	 * @return Intersection point
	 */
	public static Point getLineIntersection(
			Point c1, Point v1, Point c2, Point v2) {

		// ohne richtungsvektoren können wir nicht rechnen
		if (v1.x == 0 && v1.y == 0) return null;
		if (v2.x == 0 && v2.y == 0) return null;
		
		// Skalierung für Richtungsvektor der anderen Geraden berechnen
		// Zeit des Asteroiden bis zum Schnittpunkt
		double b = (double)(v1.x*c2.y - v1.x*c1.y - c2.x*v1.y + c1.x*v1.y) 
				 / (double)(v2.x*v1.y - v1.x*v2.y);
		
		Point intersectPoint = new Point((int) (c2.x + v2.x * b), 
										 (int) (c2.y + v2.y * b));
		
		// gleichung kann unlösbar sein - besser nichts als falschen punkt zurückgeben 
		if (intersectPoint.x == 0 && intersectPoint.y == 0) {
			intersectPoint = null;
		}
		
		return intersectPoint;
	}

	/**
	 * Calculates the distance between two Points
	 * 
	 * @param x0
	 * @param y0
	 * @param x1
	 * @param y1
	 * @return The distance.
	 */
	public static int getDistance(int x0, int y0, int x1, int y1) {
		
		Point delta = getDelta(x0, y0, x1, y1);
		int dist = (int) Math.round(Math.hypot(delta.x, delta.y));
		
		return dist;
	}

	/**
	 * Calculates the distance between two Points
	 * 
	 * @param p0 Point 1
	 * @param p1 Point 2
	 * @return The distance.
	 */
	public static int getDistance(Point p0, Point p1) {

		return getDistance(p0.x, p0.y, p1.x, p1.y);
	}

	/**
	 * Calculates the distance between two Points
	 * taking the wrapping at the borders into account
	 * 
	 * @param x0
	 * @param y0
	 * @param x1
	 * @param y1
	 * @return Distance with wrapping.
	 */
	public static Point getDelta(int x0, int y0, int x1, int y1) {

		int dx = x1 - x0;
		if (dx < -HALF_EXTENT) dx += EXTENT_X;
		if (dx >= HALF_EXTENT) dx -= EXTENT_X;

		int dy = y1 - y0;
		if (dy < -HALF_EXTENT_Y) dy += EXTENT_Y;
		if (dy >= HALF_EXTENT_Y) dy -= EXTENT_Y;
		
		return new Point(dx, dy);
	}
	
	/**
	 * Calculates the distance between two Points
	 * taking the wrapping at the borders into account
	 * 
	 * @param p0 Point 1
	 * @param p1 Point 2
	 * @return Distance with wrapping
	 */
	public static Point getDelta(Point p0, Point p1) {
		
		return getDelta(p0.x, p0.y, p1.x, p1.y);
	}

	/**
	 * Returns the shooting angle for the given rotation byte.
	 * 
	 * @param rotationByte
	 * @return The shooting angle of the ship.
	 */
	public static double getAngleFromRotByte(int rotationByte) {
		
		double angle = Math.PI / 128d * normalizeRotByte(rotationByte); 
		return normalizeAngle(angle);
	}
	
	/**
	 * Normalizes the given angle to stay to the range of [0..2PI].
	 * 
	 * @param angle The angle to normalize.
	 * @return The normalized angle.
	 */
	public static double normalizeAngle(double angle) {

		double fullTurn = 2 * Math.PI;
		
		while (angle <        0) angle += fullTurn;
		while (angle > fullTurn) angle -= fullTurn;
		
		return angle;
	}
	
	/**
	 * Normalizes a point to stay in the visible playing area.
	 * 
	 * @param p The point to normalize.
	 * @return The normalized point.
	 */
	public static Point normalizePoint(Point p) {
		
		if (p == null) return null;

		// wrap once (if neccessary)
		if (p.x < MIN_X) p.x += EXTENT_X;
		if (p.x > MAX_X) p.x -= EXTENT_X;		
		if (p.y < MIN_Y) p.y += EXTENT_Y;
		if (p.y > MAX_Y) p.y -= EXTENT_Y;
		
		// if the points are still outside the screen they are unreachable
		if (p.x < MIN_X ||
			p.x > MAX_X ||
			p.y < MIN_Y ||
			p.y > MAX_Y) 
		{
			// point is unreachable
			p = null;
		}
		
		return p;
	}
	
	/**
	 * Normalizes the rotation byte to 0..255.
	 * 
	 * @param rotByte
	 * @return
	 */
	public static int normalizeRotByte(int rotByte) {
		
		while (rotByte <   0) rotByte += 256;
		while (rotByte > 255) rotByte -= 256;
		
		return rotByte;
	}
	
	/**
	 * Predicts the rotation byte (angle) a shot will have if fired in this frame. 
	 * 
	 * @param ship
	 * @param prevFrame
	 * @return
	 */
	public static Integer getAssertedRotationByteAtShot(SpaceShip ship, 
			FrameInfo prevFrame) {

		Integer shootingRotByte = ship.getRotationByte();
		
		if (shootingRotByte != null && prevFrame != null) {
			
			if (prevFrame.wasPressedLeft()) shootingRotByte += 3;
			else if (prevFrame.wasPressedRight()) shootingRotByte -= 3;
			
			shootingRotByte = normalizeRotByte(shootingRotByte);
		}
		
		return shootingRotByte;
	}
	
//	/**
//	 * Returns the minimum distance two objects will have during their flight.
//	 * 
//	 * @param pFaster
//	 * @param vxFaster
//	 * @param vyFaster
//	 * @param pSlower
//	 * @param vxSlower
//	 * @param vySlower
//	 * @return
//	 */
//	static double getMinimumDistance(
//			Point pFaster, int vxFaster, int vyFaster, 
//			Point pSlower, int vxSlower, int vySlower) {
//		
//		if (pFaster == null || pSlower == null) return Double.POSITIVE_INFINITY;
//		
//		double a =  vyFaster - vySlower;
//		double b = -vxFaster - vySlower;
//		double c = -(a*pFaster.x + b*pFaster.y);
//		double d =  (a*pSlower.x + b*pSlower.y + c) / Math.hypot(a, b);
//		
//		return Math.abs(d);				
//	}
//	
//	/**
//	 * @see Helper#getMinimumDistance(Point, int, int, Point, int, int)
//	 * 
//	 * @param pFaster
//	 * @param vFaster
//	 * @param pSlower
//	 * @param vSlower
//	 * @return
//	 */
//	static double getMinimumDistance(
//			Point pFaster, Point vFaster, Point pSlower, Point vSlower) {
//		
//		return getMinimumDistance(
//				pFaster, vFaster.x, vFaster.y, pSlower, vSlower.x, vSlower.y);
//	}
	
	/**
	 * Get a possible aiming point for a target at a specific rotation byte.
	 * 
	 * @param ship
	 * @param target
	 * @param rotation
	 * @return The point to aim at or NULL.
	 */
	private static Point getAimingPoint(
			SpaceShip ship, Target target, int rotation) {
		
		Point aimingPoint = null;
		int targetDistance = EXTENT;
		
		// x-y-deltas vorbereiten, um torpedo-flugbahn um ziel-radius parallel 
		// zu verschieben
		double perpendicularAngle = getAngleFromRotByte(rotation) + Math.PI / 2d;
		int deltaX = (int) Math.round(
				Math.cos(perpendicularAngle) * target.getSize() * 0.5d);
		int deltaY = (int) Math.round(
				Math.sin(perpendicularAngle) * target.getSize() * 0.5d);
		
		// prepare some params we might need multiple times
		Point shipCenter = ship.getCenter();
		Point torpedoVelocity = predictTorpedoVelocity(ship, rotation);
		Point targetCenter = target.getCenter();
		Point targetVelocity = target.getVelocityVector();
		
		// erster versuch - direkte schussbahn
		Point newAimingPoint = getLineIntersection(
				shipCenter, torpedoVelocity, targetCenter, targetVelocity);
		if (newAimingPoint != null) {
			int dist = target.getDistance(newAimingPoint);
			
			if (dist < targetDistance && ship.getDistance(newAimingPoint) < TORPEDO_RANGE) {
				aimingPoint = newAimingPoint;
				targetDistance = dist;
			}
		}
		
		// zweiter versuch - eine kante des korridors
		shipCenter = new Point(ship.getX() + deltaX, ship.getY() + deltaY); 
		newAimingPoint = getLineIntersection(
				shipCenter, torpedoVelocity, targetCenter, targetVelocity);
		if (newAimingPoint != null) {
			int dist = target.getDistance(newAimingPoint);
		
			if (dist < targetDistance && ship.getDistance(newAimingPoint) < TORPEDO_RANGE) {
				aimingPoint = newAimingPoint;
				targetDistance = dist;
			}
		}

		// dritter versuch - andere kante des korridors
		shipCenter.setLocation(ship.getX() - deltaX, ship.getY() - deltaY); 
		newAimingPoint = getLineIntersection(
				shipCenter, torpedoVelocity, targetCenter, targetVelocity);
		if (newAimingPoint != null) {			
			int dist = target.getDistance(newAimingPoint);
			
			if (dist < targetDistance && ship.getDistance(newAimingPoint) < TORPEDO_RANGE) {
				aimingPoint = newAimingPoint;
				targetDistance = dist;
			}
		}

		
		// unbedingt unnormalisiert zurückgeben!!!
		return aimingPoint;
	}
	
	/**
	 * Calculates the best aiming point (alternative calculation)
	 * 
	 * @param ship Space ship
	 * @param target Target
	 * @param lastFrame The previous frame
	 * @param turningRange The max. number of steps to turn while search.
	 */
	public static void generateShootingSolution(SpaceShip ship, Target target, 
			FrameInfo lastFrame) {
		
		// large asteroids are ignored until they are perfectly tracked
		if (target instanceof Asteroid) {
			Asteroid ast = (Asteroid) target;
			if (ast.getNativeSize() == Asteroid.AsteroidSize.BIG) {
				if (!ast.isTrackingComplete()) {
//					System.out.println("großen brocken ignoriert");
					return;
				}
			}
		}
		
		// we won't find a correct point for underdeterminated targets
		if (!target.hasKnownVelocity()) return;
		
		// asserted direction our torpedo will have if shot within this frame
		Integer initialRotation = getAssertedRotationByteAtShot(ship, lastFrame);
		if (initialRotation == null) return;
		
		// the possible rotation we search a solution for 
		int rotation;
		
		// the criteria to find the best solution
		int time;
		int minTime = 9999; // time of the best solution
		int minLeftTime = 9999; // time of the best solution for left turns
		int minRightTime = 9999; // time of the best solution for right turns 
		
		// let's try some turns and test for best aiming point
		ShootingSolution newSolution = null;
		ShootingSolution shootingSolution = null;
		
		boolean leftGetsWorse = false, rightGetsWorse = false;

		// drehung -42..42 entspricht ca. -180..180°
		for (int i = 0; i < 43 + 1; i++) {
			
			if (!leftGetsWorse) {
				// check counterclockwise rotation
				rotation = normalizeRotByte(initialRotation + i * 3);
				newSolution = getShootingSolution(ship, target, rotation, i);
				
				if (newSolution != null) {
//					time = newSolution.getTotalTime();
					time = newSolution.getTurningTime() + newSolution.getWaitingTime();
					
					if (time < minTime) {
						// found better solution
						shootingSolution = newSolution;
						shootingSolution.setTurningDirection(Direction.LEFT);
						minTime = time;
						minLeftTime = time;
					} 
					else if (time > minLeftTime) {
						// no better solution for left turns
						leftGetsWorse = true;
					}
				}
			}

			// don't check twice for the 0-rotation 
			if (i == 0) continue;
			
			if (!rightGetsWorse) {
				// check clockwise rotation
				rotation = normalizeRotByte(initialRotation - i * 3);
				newSolution = getShootingSolution(ship, target, rotation, i);	
				
				if (newSolution != null) {
//					time = newSolution.getTotalTime();
					time = newSolution.getTurningTime() + newSolution.getWaitingTime();
					
					if (time < minTime) {
						// found better solution
						shootingSolution = newSolution;
						shootingSolution.setTurningDirection(Direction.RIGHT);
						minTime = time;
						minRightTime = time;
					} 
					else if (time > minRightTime) {
						// no better solutions for right turns
						rightGetsWorse = true;
					}
				}
			}
			
			// stop if search in both directions becomes worse
			if (leftGetsWorse && rightGetsWorse) {
//System.out.println("stopping search after "+i+"turns.");
				break;
			}
		}
		
		if (shootingSolution != null) {
			// solution for this target found
			if (target instanceof Asteroid) {
				Asteroid ast = (Asteroid) target;
				if (ast.getNativeSize() == Asteroid.AsteroidSize.BIG)
					shootingSolution.setShotsPlanned(4);
				else if (ast.getNativeSize() == Asteroid.AsteroidSize.MIDDLE)
					shootingSolution.setShotsPlanned(2);
			}
			else if (target instanceof Ufo) {
				shootingSolution.setShotsPlanned(2);
			}
			
			target.setShootingSolution(shootingSolution);
		}
	}

	/**
	 * Search for a shooting solution based on the given ship rotation.
	 * 
	 * @param ship The ship.
	 * @param target The target.
	 * @param rotByte The rotation to find a solution for.
	 * @param turns The number of turns needed to match this rotation.
	 * @return A shooting solution or <code>null</code>.
	 */
	private static ShootingSolution getShootingSolution(SpaceShip ship, 
			Target target, int rotByte, int turns) {

		ShootingSolution shootingSolution = null;
		
		// is there any point where we can hit the target?
		Point aimingPoint = getAimingPoint(ship, target, rotByte);
		Point wrappedPoint = normalizePoint(aimingPoint);
		
		if (aimingPoint == null || wrappedPoint == null) {
			// point is more than one screen wrap away - unreachable
			//System.out.println("keinen schnittpunkt gefunden");			
		}
		else if (!ship.isPointAhead(aimingPoint, rotByte)) {
			// the point is not ahead of the ship with the future rotation
		}
		else if (!target.isPointAhead(aimingPoint)) {
			// aimingpoint is behind the target
		}
		else {
			// how long will a torpedo need to reach this point?
			// 0.5 frames von der flugzeit abziehen, weil das torpedo 2 frames nach 
			// dem schuss in 2.5 torpedo-speeds entfernung von der schiffsmitte 
			// auftaucht
			int torpedoTime = (int) Math.round(
					(double) ship.getDistance(aimingPoint) 
					/ (double) TORPEDO_SPEED - .5d);
			
			// how long will the target need to reach this point?
			int targetTime = (int) Math.round(
					(double) target.getDistance(aimingPoint) 
					/ (double) target.getVelocity());
			
			// how long do we have to wait for this solution?
			int waitingTime = targetTime - torpedoTime - turns;
			
			if (torpedoTime > TORPEDO_LIFETIME) {
				// point is too far away
				System.err.println("oO! -> torpedoTime="+torpedoTime);
			}
			else if (waitingTime < 0 || waitingTime > 10) {
				// solution is unsolvable or too far in the future
			}
			else {
				// we have a solution
				shootingSolution = new ShootingSolution(target.getUid(), 
						wrappedPoint, rotByte, turns, waitingTime, torpedoTime);
			}
		}
		
		return shootingSolution;
	}
	
	/**
	 * Checks all possible targets if they should have been destroyed but are
	 * still alive - marks them as alive, so they can be shot again.
	 * 
	 * @param currFrame The current frame information.
	 * @param prevFrame The previous frame information.
	 */
	public static void handleZombies(FrameInfo currFrame, FrameInfo prevFrame) {
		
		/*
		 * make a list of targets that are still under fire to ensure, that shot
		 * markers are removed only if there is no bullet left aiming on this
		 * target
		 */
		ArrayList<Integer> targetUnderFire = new ArrayList<Integer>();
		
		/*
		 * check for each bullet if distance to it's target is less than in the 
		 * previous frame - otherwise it missed, so remove dead-flag from target
		 */
		for (Bullet currBullet: currFrame.getBullets()) {
			
			GameObject o;
			Bullet prevBullet;
			Target prevTarget;
			Target currTarget;
			
			// find this bullet in the prev frame
			o = prevFrame.getObjectByUid(currBullet.getUid());
			if (o == null) continue;
			else prevBullet = (Bullet) o;
			
			// find it's target in the prev frame
			o = prevFrame.getObjectByUid(prevBullet.getTargetUid());
			if (o == null) continue;
			else prevTarget = (Target) o; 
			
			// find it's target in the curr frame
			o = currFrame.getObjectByUid(currBullet.getTargetUid());
			if (o == null) continue;
			else currTarget = (Target) o;
			
			int prevDist = prevBullet.getDistance(prevTarget);
			int currDist = currBullet.getDistance(currTarget);
			
			if (currDist < prevDist) {
				// torpedo is still hot
				targetUnderFire.add(currTarget.getUid());
			}
		}
		
		for (Target t: currFrame.getAllTargets()) {
			
			if (!t.isUnderFire()) continue;

			/*
			 * if the bullet that should destroy a target just disappears - it may
			 * have hit another target - so mark this one as alive again.
			 */
			Bullet bullet = null;
			// find it's designated bullet
			for (Bullet b: currFrame.getBullets()) {
				if (b.getTargetUid() == t.getUid()) {
					bullet = b;
					break;
				}
			}
			// if not found - the bullet destroyed another target
			if (bullet != null) {
				targetUnderFire.add(t.getUid());
			}
		}
		
		// if there is no bullet left that aims on this target - remove shot marker
		for (Target t: currFrame.getAllTargets()) {
			
			/*
			 * if the shot-marker is older than a torpedo lifetime, the bullet
			 * must have missed
			 */
			int shotMarkerAge = currFrame.getFrameId() - t.getMarkedAsShotAtFrame(); 

			if (!targetUnderFire.contains(t.getUid()) 
					|| shotMarkerAge > TORPEDO_LIFETIME) {
				t.removeMarkedAsShot();
			}
		}
	}
	
	/**
	 * Get the assumed torpedo velocity for a given shot-angle.
	 * 
	 * @param angle The angle of the ship at shooting time.
	 * @param ship 
	 * @return 
	 */
	public static Point predictTorpedoVelocity(SpaceShip ship, double angle) {
		
		int vx = (int) Math.round(Math.cos(angle) * TORPEDO_SPEED) + ship.getVelocityX();
		int vy = (int) Math.round(Math.sin(angle) * TORPEDO_SPEED) + ship.getVelocityY();
		
		return new Point(vx, vy);
	}
	
	/**
	 * @see Helper#predictTorpedoVelocity(double)
	 * 
	 * @param rotationByte The rotation byte of the ship at shooting time.
	 * @param ship
	 * @return
	 */
	public static Point predictTorpedoVelocity(SpaceShip ship, int rotationByte) {

		double angle = getAngleFromRotByte(rotationByte);
		
		return predictTorpedoVelocity(ship, angle);
	}
	
//	/**
//	 * Returns TRUE if we have a shot left for a secondary target. 
//	 * 
//	 * @param frame The current frame info.
//	 * @return True if there is a shot left.
//	 */
//	public static boolean secondaryShotAvailable(FrameInfo frame) {
//		
//		boolean freeShot = false;
//		
//		if (frame.getShootingSolution() == null) {
//			// if we have no primary target, there is no reason not to shoot
//			freeShot = true;
//		}
//		else if (frame.getOwnBulletCount() < 3) {
//			// if we have two shots left - just do it
//			freeShot = true;
//		}
//		else if (frame.getOwnBulletCount() < 4) {
//			// we only have one shot left at the moment
//			// take the time we need to reach the correct rotation for the primary target
//			int timeLeft = frame.getShootingSolution().getTurningTime();
//			
//			// check each bullet if it might disappear before we plan to shoot 
//			for (Bullet b: frame.getBullets()) {
//				
//				if (b.getShootingSolution() == null) {
//					// bullet is too young and does not yet know its solution
//					continue;
//				}
//				
//				// get the time this bullet has left to live
//				int timeToLive = b.getShootingSolution().getShootingTime() > b.getAge()
//						// this might be the time till it hits it's target
//						? b.getShootingSolution().getShootingTime() - b.getAge()
//						// or if it missed the time until it dies
//						: TORPEDO_LIFETIME - b.getAge();
//				if (timeToLive < timeLeft) {
//					// torpedo will die before we need the next one - shoot!
//					freeShot = true;
////System.out.println("secondaryShotAvailable: we need to shoot in "+timeLeft+" frames, a bullet will die in "+timeToLive+" frames - OK");					
//					break;
//				}
//			}
//		}
//		
//		return freeShot;
//	}
}

// TODO torpedo-statistik: wie viele, treffer, fehlschüsse
// TODO ziele werden nicht mehr grün und auch nciht mehr blau!
 