/**
 * 
 */
package de.zappbrannigan.ai;

import java.util.Vector;

import de.zappbrannigan.communication.FrameInfo;
import de.zappbrannigan.objects.*;

/**
 * A class that selects the current target for the GromitBonsaiAsteroidPlayer 
 * based on evidence, worth or danger.
 * 
 * @author gromit
 */
public class TargetSelector implements GameData {

	/** The screen center. */
	private DummyTarget centerTarget;
	
	/**
	 * Constructor
	 */
	public TargetSelector() {
		
		// create the center target
		centerTarget = new DummyTarget(HALF_EXTENT, HALF_EXTENT);
		// and assign a uid  
		centerTarget.setUid(AsteroidPool.getAsteroidPool().getNextObjectUid());
	}
	
	/**
	 * Returns a target (if any) that may collide with the ship. If there are
	 * more than one, the first arriving is chosen.
	 * 
	 * @param ship
	 * @param frame
	 * @return A dangerous target
	 */
	public static Target findDangerousTarget(FrameInfo frame) {
		
		// abort if no ship is present
		if (!frame.hasShip()) return null;
		
		Target mostDangerousTarget = null;
		SpaceShip ship = frame.getSpaceShip();
		
		// sinnvolle ziele sollten in torpedo-reichweite sein
		double minImpactTime = TORPEDO_LIFETIME;

		// ufo darf nicht wichtiger sein, als ein bald einschlagender asteroid
		// deshalb per default auswählen - und ggf. mit asteroiden überschreiben
		if (frame.hasUfo() && !frame.getUfo().isUnderFire()) {
			mostDangerousTarget = frame.getUfo();
		}

		// nun die asteroiden testen
		for (Target ast : frame.getAsteroids()) {
			
			// wenn schon beschossen - ignorieren
			if (ast.isUnderFire()) continue;
			
			/* 
			 * wenn abstand zur flugbahn > n, dann ist das objekt ungefährlich
			 * n = schiffsgröße + objektgröße + sicherheitsreserve
			 */
			if (ship.getFlyByDistance(ast) > ship.getSize() + ast.getSize() + 8)
				continue;
			
			// wenn sich das objekt entfernt - auch ignorieren
			if (ship.getTargetDirection(ast) > 0) continue;
			
			// als "gefährder" markieren
			ast.setCritical(true);
			
			/*
			 * anzahl frames bis zum einschlag berechnen - entscheidend ist 
			 * nicht die entfernung, sondern welcher mich zuerst trifft
			 */
			double impactTime = 
				(ship.getDistance(ast) - ship.getSize() - ast.getSize()) 
				/ ast.getVelocity();
			// time we need to turn into target direction - astimation
			double turningTime = ship.getDeltaAngle(ast) / (Math.PI / 43d);
			// shorten the impact time by the time we need to turn and shoot
			impactTime -= turningTime;
			
			if (impactTime < minImpactTime) {
				minImpactTime = impactTime;
				mostDangerousTarget = ast;
			}
		}
		
		return mostDangerousTarget;
	}
	
	/**
	 * Find the most interesting target - this method is just used, if no other
	 * potentially dangerous objects are left.
	 * 
	 * @param ship ship info
	 * @param frame frame info
	 * @return nearest target or <code>null</code> if there is no target near
	 *         enough
	 */
	public static Target findValuableTarget(FrameInfo frame) {

		Target mostValuableTarget = null;
		Vector<Asteroid> candidates = new Vector<Asteroid>();
		
		// now make a list of possible new targets
		for (Asteroid ast : frame.getAsteroids()) {
			
			// without shooting solution we can't evaluate the target
			if (!ast.hasShootingSolution()) {
				continue;
			}

			// ignore dying targets
			if (ast.isUnderFire()) {
				continue;
			}
			
			// ignore if the asteroids array is full and we would loose points
			if (ast.getNativeSize() != Asteroid.AsteroidSize.SMALL && 
					frame.getAsteroids().size() > 24) {
					System.out.println("ziel ignoriert wegen vollem array!");
				continue;
			}
			
			candidates.add(ast);
		}
		
		if (candidates.size() == 1) {
			// if there is only one target - don't fuzzle around, just use it
			mostValuableTarget = candidates.firstElement();
		}
		else if (!candidates.isEmpty()) {
//System.out.println(candidates.size()+"/"+frame.getAsteroids().size()+" have shooting solutions.");
			mostValuableTarget = findFirstDyingTarget(frame, candidates);
//			mostValuableTarget = findYoungestTarget(frame, candidates);
//			mostValuableTarget = findValuableTargetWithCostFunction(frame, candidates);
//			mostValuableTarget = findValuableTargetWithPriorityList(frame, candidates);
		}
		
		Ufo ufo = frame.getUfo();
		if (ufo != null && !ufo.isUnderFire()) {
			// egal, wie weit weg - ufo immer zuerst abschiessen
			// wenn es aber als abgeschossen markiert ist, kann man es erstmal
			// ignorieren
			mostValuableTarget = ufo;
		}
		// System.out.println("Nearest: "+nearest);
		return mostValuableTarget;
	}

	/**
	 * First strategy with cost function
	 * 
	 * @param info
	 * @param possibleAsteroids
	 * @return
	 */
	private static Target findValuableTargetWithCostFunction(FrameInfo info, Vector<Asteroid> possibleAsteroids) {
		
		Target mostValueable = null;
		
		SpaceShip ship = info.getSpaceShip();
		if (ship != null) {
			double maxRoS = 0; // return on shot a.k.a. kostenfunktion
			
			for (Asteroid ast : possibleAsteroids) {
				
				// TODO try total time instead of turning time
				// the shorter the total time the earlier bullets are
				// available again
				// ignore included points, we will get them anyway
				
	//			double ros = (double) ast.getPoints() / (double) ast.getShootingSolution().getTurningTime(); 
				double ros = ((double) ast.getPoints() + 0.5 * (double)ast.getIncludedPoints()) / 
								(double) ast.getShootingSolution().getTurningTime(); 
				
				if (ros > maxRoS) { 
					maxRoS = ros;
							
					mostValueable = ast;
				}
			}
		}
		
		return mostValueable;
	}
	
	/**
	 * First strategy with cost function
	 * 
	 * @param info
	 * @param possibleAsteroids
	 * @return
	 */
	private static Target findFirstDyingTarget(FrameInfo info, Vector<Asteroid> possibleAsteroids) {
		
		Target mostValueable = null;
		
		int time;
		int minTime = 9999;
		
		for (Asteroid target: possibleAsteroids) {
			// criteria is the time until the target dies
//			time = target.getShootingSolution().getTotalTime();
			// criteria is the time until the shot is ignited
			time = target.getShootingSolution().getTurningTime() 
					+ target.getShootingSolution().getWaitingTime();
			
			if (time < minTime) {
				
				minTime = time;
				mostValueable = target;
			}
		}
		
		return mostValueable;
	}
	
	/**
	 * First strategy with cost function
	 * 
	 * @param info
	 * @param possibleAsteroids
	 * @return
	 */
	private static Target findYoungestTarget(FrameInfo info, Vector<Asteroid> possibleAsteroids) {
		
		Target target = null;
		
		double minAge = 1000;
		
		for (Asteroid a: possibleAsteroids) {
			
			if (a.getAge() < minAge) {
				
				minAge = a.getAge();
				target = a;
			}
		}
		
		return target;
	}
	
	/**
	 * Second strategy with priority lists
	 * 
	 * @param ship
	 * @param info
	 * @param possibleAsteroids
	 * @return
	 */
//	private static Target findValuableTargetWithPriorityList(SpaceShip ship, FrameInfo info, Vector<Asteroid> possibleAsteroids) {
//		
//		Target mostValueable = null;
//		Vector<Asteroid> currVector = null;
//		Vector<Asteroid> leftVector = new Vector<Asteroid>();
//		Vector<Asteroid> rightVector = new Vector<Asteroid>();
//
////		System.out.println(possibleAsteroids.size() + " possibleAsteroids found");
//		int count = 0;
//		for (Asteroid ast : info.getAsteroids()) {
//		
//			if (possibleAsteroids.indexOf(ast) > -1) {
//				
//				count++;
//			}
//		}
////		System.out.println(count + " possibleAsteroids found again");
//		
//		return mostValueable;
//	}
	
	/**
	 * Returns a torpedo that will hit the ship.
	 * 
	 * @param ship
	 * @param frame
	 * @return A dangerous torpedo or <code>NULL</code> if there is none.
	 */
	public static Bullet findCollidingBullet(SpaceShip ship, FrameInfo frame) {

		Bullet bullet = null;
		
		for (Bullet b: frame.getBullets()) {
			if (ship.getDistance(b.getFuturePosition(2)) < ship.getSize()) {
				
				bullet = b;
				break;
			}
		}
		
		return bullet;
	}
	
	/**
	 * Returns an asteroid that will hit the ship.
	 * 
	 * @param ship
	 * @param frame
	 * @return An colliding asteroid or <code>NULL</code>.
	 */
	public static Asteroid findCollidingAsteroid(SpaceShip ship, FrameInfo frame) {

		Asteroid asteroid = null;
		
		for (Asteroid a: frame.getAsteroids()) {
			if (ship.getDistance(a.getFuturePosition(2)) < a.getSize() + ship.getSize()) {
				
				asteroid = a;
				break;
			}
		}
		
		return asteroid;
	}
	
	/**
	 * Returns a dummy target at the screen center.
	 * 
	 * @return Dummy target at screen center.
	 */
	public Target getCenterTarget() {
		
		return centerTarget;
	}
	
//	/**
//	 * Returns true, if the target is identical to the center target.
//	 * 
//	 * @param t The target to compare.
//	 * @return True if the target is the center target.
//	 */
//	public boolean isCenterTarget(Target t)
//	{
//		
//		return (t.equals(centerTarget));
//	}

}
