// ============================================================================
// File:               $File$
//
// Project:            
//
// Purpose:            
//
// Author:             Rammi
//
// Copyright Notice:   (c) 2008  Rammi (rammi@caff.de)
//                     This code is in the public domain.
//                     Use at own risk.
//                     No guarantees given.
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.ai;

import java.awt.Point;
import java.util.*;

import de.zappbrannigan.communication.FrameInfo;
import de.zappbrannigan.communication.FramePreparer;
import de.zappbrannigan.objects.Asteroid;
import de.zappbrannigan.objects.Bullet;
import de.zappbrannigan.objects.SpaceShip;
import de.zappbrannigan.objects.Ufo;

/**
 * Helper class which tries to add velocities to game objects. This is done in a
 * too simple-minded way by just comparing the last two screens.
 * 
 * This class is part of a solution for a
 * <a href="http://www.heise.de/ct/creativ/08/02/details/">competition by the
 * German computer magazine c't</a>.
 */
public class VelocityPreparer implements FramePreparer {
	
	/** overall frame counter - counts every frame, used for time measurement */
	private int frameCount = 0;
	
	/**
	 * Helper class to presort asteroids to allow easier compare between frames.
	 */
	private static class AsteroidSelector {
		/** Maximum distance we assume an asteroid may move between frames. */
		private static final int MAX_DIST = 80;

		/**
		 * Helper class: key for mapping asteroids.
		 */
		private static class Key {

			/** The asteroid's type. */
			private final int type;
			/** The asteroids size. */
			private final int size;

			/**
			 * Constructor.
			 * 
			 * @param ast asteroid
			 */
			private Key(Asteroid ast) {

				this.type = ast.getType();
				this.size = ast.getSize();
			}

			/**
			 * @see Object#equals(Object)
			 */
			@Override
			public boolean equals(Object o) {

				if (this == o) {
					return true;
				}
				if (o == null || getClass() != o.getClass()) {
					return false;
				}

				Key key = (Key) o;

				if (size != key.size) {
					return false;
				}
				if (type != key.type) {
					return false;
				}

				return true;
			}

			/**
			 * @see Object#hashCode()
			 */
			@Override
			public int hashCode() {

				int result;
				result = type;
				result = 31 * result + size;
				return result;
			}
		}

		/** Presorted mapping of asteroids. */
		private Map<Key, Collection<Asteroid>> sorted = new HashMap<Key, Collection<Asteroid>>();

		/**
		 * Constructor.
		 * 
		 * @param asteroids asteroids to be presorted
		 */
		private AsteroidSelector(Collection<Asteroid> asteroids) {

			for (Asteroid ast : asteroids) {
				Key key = new Key(ast);
				Collection<Asteroid> list = sorted.get(key);
				if (list == null) {
					list = new LinkedList<Asteroid>();
					sorted.put(key, list);
				}
				list.add(ast);
			}
		}

		/**
		 * Get the best matching asteroid assuming the keys collected here are
		 * from the previous frame while the given asteroid is from the current
		 * frame.
		 * 
		 * @param asteroid previous frame asteroid
		 * @return best match or <code>null</code>
		 */
		public Asteroid getBestMatch(Asteroid asteroid) {

			Key key = new Key(asteroid);
			Collection<Asteroid> list = sorted.get(key);
			SortedMap<Integer, Asteroid> result = new TreeMap<Integer, Asteroid>();

			if (list != null) {
				Point futurePosition = asteroid.getFuturePosition(1);
				for (Asteroid a : list) {
					int dist2 = a.getDistance(futurePosition);
					if (dist2 < MAX_DIST) {
						result.put(dist2, a);
					}
				}
			}
			return result.isEmpty() ? null : result.values().iterator().next();
		}
	}

	/**
	 * Helper class to presort bullets to allow easier compare between frames.
	 */
	private static class BulletSelector {

		/** Maximum distance we assume a bullet may move between frames. */
		private static final int MAX_DIST = 100;

		Collection<Bullet> listOfBullets;

		/**
		 * Constructor.
		 * 
		 * @param asteroids asteroids to be presorted
		 */
		private BulletSelector(Collection<Bullet> bullets) {

			listOfBullets = bullets;
		}

		/**
		 * Get the best matching bullet
		 * 
		 * @param bullet Previous frame bullet
		 * @return best match or <code>null</code>
		 */
		public Bullet getBestMatch(Bullet bullet) {
			/*
			 * TODO bullets werden teilweise nicht wiedererkannt, wenn sie über 
			 * den rand fliegen!
			 */
			Bullet bestMatch = null;
			Point futurePosition = bullet.getFuturePosition(1);
			double minDistance = MAX_DIST;

			for (Bullet currBullet : listOfBullets) {

				double distance = currBullet.getDistance(futurePosition);

				if (distance < minDistance) {

					bestMatch = currBullet;
					minDistance = distance;
				}
			}

			return bestMatch;
		}
	}

	/**
	 * Prepare the frame(s).
	 * 
	 * @param frameInfos the collected frame infos
	 */
	public void prepareFrames(LinkedList<FrameInfo> frameInfos) {

		if (frameInfos.size() > 1) {
			FrameInfo currFrame = frameInfos.getLast();
			FrameInfo prevFrame = frameInfos.get(frameInfos.size() - 2);
			
			updateFrameCounter(currFrame, prevFrame);

			// update ufo
			Ufo ufo = currFrame.getUfo();
			if (ufo != null && prevFrame.hasUfo()) {
				ufo.merge(prevFrame.getUfo(), frameInfos);
			}

			// update ship
			SpaceShip ship = currFrame.getSpaceShip();
			if (ship != null && prevFrame.hasShip()) {
				ship.merge(prevFrame.getSpaceShip(), frameInfos);
			}
			
			// update asteroids
			AsteroidSelector selector = new AsteroidSelector(currFrame.getAsteroids());
			for (Asteroid ast : prevFrame.getAsteroids()) {
				Asteroid candidate = selector.getBestMatch(ast);
				if (candidate != null) {
					candidate.merge(ast, frameInfos);
				}
			}
 			
			BulletSelector bselector = new BulletSelector(currFrame.getBullets());
			for (Bullet bullet : prevFrame.getBullets()) {
				Bullet candidate = bselector.getBestMatch(bullet);
				if (candidate != null) {
					candidate.merge(bullet, frameInfos);
				}
			}

			// handle missed targets
			Helper.handleZombies(currFrame, prevFrame);
		}
	}
	
	/**
	 * Increments the frame counter.
	 * 
	 * @param currFrame The current frame.
	 * @param prevFrame The previous frame.
	 */
	public void updateFrameCounter(FrameInfo currFrame, FrameInfo prevFrame) {
		
		if (frameCount == 1 && !currFrame.hasAsteroids()) {
			// the game has not yet begun
		}
		else {
			// detect new game
			if (prevFrame.getLives() == 0 && currFrame.getLives() == 3) {
				// reset frame counter
				frameCount = 0;
			}

			// tag frame and increment counter
			currFrame.setFrameId(frameCount++);
		}
	}
}
