// ============================================================================
// File:               $File$
//
// Project:            
//
// Purpose:            
//
// Author:             Harald Boegeholz / Rammi
//
// Copyright Notice:   (c) Original C++ code: Harald Boegeholz (c't)
//                     (c) Java port: 2008  Rammi (rammi@caff.de)
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.ai;

import java.io.IOException;

//import org.apache.log4j.Logger;

import de.zappbrannigan.communication.*;
import de.zappbrannigan.objects.*;

/**
 * Sample player quite similar to the one programmed by Harald Boegeholz.
 * 
 * It does not use threading, but instead calculates everything in the
 * frameReceived() method and so runs in the communication thread. Usually it's
 * preferable to make the calculations in an own thread, so the communication
 * thread is just doing communication.
 * 
 * This class is part of a solution for a <a
 * href="http://www.heise.de/ct/creativ/08/02/details/">competition by the
 * German computer magazine c't</a>.
 */
public class ZappBranniganAsteroidPlayer 
		extends AsteroidPlayer 
		implements FrameListener, GameData {
	
//	private static final Logger log = Logger.getLogger(ZappBranniganAsteroidPlayer.class);

	/** The number of frames to remember. */
	private final int FRAME_BUFFER_LENGTH = 10;
	/** The area around screen center where our ship should stay. */
	private final int HOME_AREA_RADIUS = 200;
	/** thinking time over all frames */
	private long overallThinkTime = 0;
	/** Holds all keys pressed by the player */
	private byte keys = 0;
	/** The number of lives in this game. */
	private int lives = 0;
	/** The game score */
	private int score = 0;
	/** Number of 100k overflows. */
	private int scoreOverFlows = 0;
	/** Number of minutes played. */
	private int gameMinutes = 0;	
	/** Limit the game to 5 minutes. */
	private final boolean limitGameTime = true;
	/** Indicates that the limited game time is over. */
	private boolean gameOver = false;
	/** The (hopefully) correct internal rotation of the ship. */
	private int rotationByte = 0;
	/** Statistics: Counter for wrong rotation bytes. */
	private long badRotationBytes = 0;
	/** The one that selects our next targets. */
	private TargetSelector targetSelector;
	/** The plan to destroy the next target. */
	private ShootingSolution shootingPlan;
	
//	private HashMap<Integer, Bullet> angleBytes = new HashMap<Integer, Bullet>();
	
	/**
	 * Constructor.
	 * 
	 * @param com
	 *            communications object for key sending
	 */
	public ZappBranniganAsteroidPlayer(Communication com) {
		super(com);
		com.setFramePreparer(new VelocityPreparer());
		targetSelector = new TargetSelector();
		initializeVariables();
	}

	/**
	 * Initialize all variables
	 */
	private void initializeVariables() {

		frames.clear();
		incomingFrames.clear();
		overallThinkTime = 0;
		score = 0;
		scoreOverFlows = 0;
		gameMinutes = 0;
		gameOver = false;
		badRotationBytes = 0;
	}
	
	/**
	 * Called each time a frame is received.
	 * 
	 * <b>ATTENTION:</b> this is called from the communication thread!
	 * Implementing classes must be aware of this and take care by
	 * synchronization or similar!
	 * 
	 * @param frame
	 *            the received frame
	 */
	@Override
	public void frameReceived(FrameInfo frame) {
		
		if (frame == null) return;
		
		// handle game score
		handleScoreCount(frame);
		// handle live count
		handleLiveCount(frame);
		// show status and limit game time
		showGameStatus(frame);

		// we can only act if our ship is present
		if (frame.hasShip() && !gameOver && frames.size() > 1)  {
			
			// initially no keys are pressed
			keys = 0;
			// update the ships rotation byte
			updateRotationByte(frame);
			// update target shooting solutions
			updateShootingSolutions(frame);
			// find a new shooting plan if neccessary
			updateShootingPlan(frame);
			// shoot
			shoot(frame);
			// if a shot was fired we can already turn to the next target
			updateShootingPlan(frame);
			// turn
			turn(frame);
			// fly around - or better not?
			accelerate(frame);
			// save us from colliding objects
			rescue(frame);
		} 
		else {
			
			// try to start a new game
			keys = BUTTON_START;
		}
		
		// send keys every time
		try {
			com.sendKeys(keys);
			frame.setPressedKeys(keys);
		} catch (IOException e) {
			e.printStackTrace();
		}

		// add current frame to the list
		frames.add(frame);
		// and care about the frame buffer
		if (frames.size() > FRAME_BUFFER_LENGTH) {
			// drop oldest frame from buffer
			FrameInfo oldFrame = frames.removeFirst();
            // and return objects to the pool
            for (Asteroid ast : oldFrame.getAsteroids()) {
            	AsteroidPool.getAsteroidPool().releaseAsteroid(ast);
            }
		}
		
		// Zeit zusammenrechnen
		overallThinkTime += System.currentTimeMillis() - frame.getReceiveTime(); 
	}
	
	/**
	 * Press hyper space key only once
	 * 
	 * @param reason Reason for pressing hyper space button
	 */
	private void pressHyperSpace(FrameInfo frame, String reason) {
		
		if (frames.getLast() != null && !frames.getLast().wasPressedHyperspace()) {
			// push da button
			keys = BUTTON_HYPERSPACE;
			
			//printTimeStamp(frame);
			System.out.println("Hyperspace wegen " + reason);
		}
	}
	
	/**
	 * Press fire
	 */
	private void pressFire() {
		
		keys |= BUTTON_FIRE;
	}
	
	/**
	 * Update live counter
	 */
	private void handleLiveCount(FrameInfo frame) {
		
		if (lives == 0 && frame.getLives() > 0) {			
			System.out.println("===> New game <===");
			handleNewGame();
		} 
		else if (lives < frame.getLives()) {
			printTimeStamp(frame);
			System.out.println("Got new live!");
		}
		else if (lives > frame.getLives()) {
			
			printTimeStamp(frame);
			System.out.println("DIED!!! :-/");
		}
		lives = frame.getLives();
	}
	
	/**
	 * Update game score and handle the 100000-points overflow.
	 * 
	 * @param The newly received frame.
	 */
	private void handleScoreCount(FrameInfo frame) {
		
		// TODO overflow doesn't work anymore
		if (lives > 0 && frame.getScore() < score % 100000) {
			// score overflow
			scoreOverFlows++;
			System.out.println("Yeah, we got an overflow: " + score);
			System.out.println("Time needed: " + getTimeStamp(frame));
		}
		
		if (frame.getScore() + scoreOverFlows * 100000 > score) {
			score = frame.getScore() + scoreOverFlows * 100000;
		}
	}
	
	/**
	 * Handles new game event
	 */
	private void handleNewGame() {
		
		initializeVariables();
	}
	
	/**
	 * Returns the time since game start as formatted string.
	 * 
	 * @return Time since game start.
	 */
	private String getTimeStamp(FrameInfo frame) {
		
		return String.format("%02d:%02d", (frame.getFrameId()-1) / 3600, 
				(frame.getFrameId()-1)/60 % 60);
	}
	
	private void printTimeStamp(FrameInfo frame) {
		
		System.out.print(
				"[Time: " + getTimeStamp(frame) + 
				", Score: " + score + 
				", Lives: " + lives + "] \n");
	}
	
	/**
	 * Updates the ships rotationByte for the current frame.
	 * 
	 * @param ship The space ship.
	 */
	private void updateRotationByte(FrameInfo currFrame) {
		
		// letzter frame, um nicht-drehung zu bemerken
		FrameInfo prevFrame = frames.getLast();
		// vorletzter frame, um zu wissen, ob drehen verlangt wurde
		FrameInfo keyFrame = frames.get(frames.size() - 2);
		// it won't work without a ship in this and the previous frame
		if (!currFrame.hasShip() || !prevFrame.hasShip()) return;

		SpaceShip ship = currFrame.getSpaceShip();
		// scheinbare schiffsrichtung merken - atan berechnen ist teuer
		double shipDirectionAngle = ship.getDirectionAngle();
		
		Integer newRotation = null;
		
		/*
		 * if we have the luck of a 270°/810° direction -> sync the 
		 * rotationByte. these are the only two positions where internal
		 * and visible rotation of the ship exactly match. 
		 */
		if (shipDirectionAngle == Math.PI/2) {
			newRotation = 64;
		} 
		else if (shipDirectionAngle == -Math.PI/2) {
			newRotation = 192;
		}
		else if (shipDirectionAngle == prevFrame.getSpaceShip().getDirectionAngle()
				&& shipDirectionAngle != 0 && shipDirectionAngle != Math.PI) {
			/*
			 * winkelbyte-korrektur anhand der punkte, an denen man sich einen
			 * schritt weiterdrehen kann, ohne dass das schiff sich bewegt:
			 */
			if (keyFrame.wasPressedLeft() || keyFrame.wasPressedRight()) {
				// passt! jetzt nach einem passend rotationByte für die aktuelle richtung suchen
				for (int b = 0; b < 256; b++) {
					if (AngleByteHashMap.getRotation(b) != shipDirectionAngle) {
						continue;
					}
					
					// und nun prüfen, ob das nächste winkelbyte auch eine nicht-drehung ergeben würde
					if (AngleByteHashMap.getRotation(b) != AngleByteHashMap.getRotation(b + 3)) {
						continue;
					}
					
					// gefunden - jetzt müssen wir nur noch bestimmen, ob das linkere oder rechtere das korrekte ist
					if (keyFrame.wasPressedLeft()) {
						newRotation = b + 3;
						break;
					}
					else if (keyFrame.wasPressedRight()) {
						newRotation = b;
						break;
					}
				}
			}
		}
		else {
			/*
			 * fallback-lösung - einfach das letzte rot-byte mit gedrückten tasten abgleichen
			 */
			if (keyFrame.wasPressedLeft()) {
				newRotation = rotationByte + 3;
			}
			else if (keyFrame.wasPressedRight()) {
				newRotation = rotationByte - 3;
			}
		}
		
		if (newRotation != null) {
			
			rotationByte = newRotation;
			// handle overflows
			rotationByte = Helper.normalizeRotByte(rotationByte);
			
			/*
			 * debugging - fehler erkennen, sollte aber theoretisch nur noch 
			 * vorkommen, wenn das schiff direkt nach rechts oder links schaut, 
			 * da dort 2-3 punkte existieren, welche der lost-step-regel entsprechen
			 * aber nicht genauer unterschieden werden können.
			 */
			if (!AngleByteHashMap.isRotationCorrect(rotationByte, ship.getDirectionAngle())) {
				
				badRotationBytes++;
	//			System.out.println("bad rotation byte: "+rotationByte);
			}
			
			ship.setRotationByte(rotationByte);
		}
	}

	/**
	 * Set a new shooting plan.
	 * 
	 * @param shootingPlan the shootingPlan to set
	 */
	private void setShootingPlan(ShootingSolution shootingPlan) {
		
		if (shootingPlan != null) {
			this.shootingPlan = shootingPlan;
			//System.out.println("setting new shooting plan on target#"+shootingPlan.getTargetUid());		
		}
	}
	
	/**
	 * Removes the current shooting plan.
	 */
	private void invalidateShootingPlan(FrameInfo frame) {
//System.out.println("invalidating plan on target#"+shootingPlan.getTargetUid());		
		Target target = (Target) frame.getObjectByUid(shootingPlan.getTargetUid());
		if (target != null) {
			target.setShootingSolution(null);
		}
		this.shootingPlan = null;
	}
	
	/**
	 * Returns TRUE if we have a shooting plan.
	 * 
	 * @return True if there is one.
	 */
	private boolean hasShootingPlan() {
		
		return (shootingPlan != null);
	}
	
	/**
	 * Get a previous frame.
	 * 
	 * @param count Number of steps into the past.
	 * @return A previous frame.
	 */
	private FrameInfo getPreviousFrame(int count) {
		
		FrameInfo frame = null;
		
		if (frames.size() >= count) {
			frame = frames.get(frames.size() - count);
		}
			
		return frame;
	}
	
	/**
	 * Get the previous frame.
	 * 
	 * @return The previous frame.
	 */
	private FrameInfo getPreviousFrame() {
		
		return getPreviousFrame(1);
	}
	
	/**
	 * Returns TRUE if a target with the given ID exists. 
	 * 
	 * @param frame
	 * @param targetUid
	 * @return
	 */
	private boolean isTargetPresent(FrameInfo frame, int targetUid) {
	
		return (frame.getObjectByUid(targetUid) != null);
	}
	
	/**
	 * Check current shooting plan and find a new one if neccessary.
	 * 
	 * @param frame
	 */
	private void updateShootingPlan(FrameInfo currFrame) {
		
		// first check, if we have a plan - but no target!
		if (!hasShootingPlan()) {
			// there is nothing to check
		}
		else if (!isTargetPresent(currFrame, shootingPlan.getTargetUid())) {
			// forget this plan
			invalidateShootingPlan(currFrame);
		}
		else {
			// is the current solution still solvable?
			Target target = (Target) currFrame.getObjectByUid(shootingPlan.getTargetUid());
		
			if (!target.isPointAhead(shootingPlan.getHitPoint())) {
//				System.out.println("cancelling plan on target#"+target.getUid());				
				invalidateShootingPlan(currFrame);
			}
		}

		Target newTarget = null;
		
		// always check for colliding/dangerous targets
		newTarget = TargetSelector.findDangerousTarget(currFrame); 
		if (newTarget != null) {
			if (!hasShootingPlan() || newTarget.getUid() != shootingPlan.getTargetUid()) {
				// override our current plan to destroy it
				if (newTarget.getShootingSolution() != null) {
					setShootingPlan(newTarget.getShootingSolution());
				}
			}
		}
		
		// make a new plan if there is no other
		if (!hasShootingPlan()) {
			// find a new target
			newTarget = TargetSelector.findValuableTarget(currFrame);
			if (newTarget != null) {
//System.out.println("new shooting plan on target #"+newTarget.getUid());
				setShootingPlan(newTarget.getShootingSolution());
			}
		}

		// if we have any plan
		if (hasShootingPlan()) {
			// indicate it's current target
			Target target = 
				(Target) currFrame.getObjectByUid(shootingPlan.getTargetUid()); 
			if (target != null) {
				// to prevent NullPointerException
				target.setCurrentTarget(true);
				// and store it in the frame object
				currFrame.setShootingSolution(shootingPlan);
			}
		}
	}
	
	/**
	 * Display the game status every minute.
	 * 
	 * @param frame
	 */
	private void showGameStatus(FrameInfo frame) {
		
		if (lives > 0 && frame.getFrameId() / 3600 > gameMinutes) {

			gameMinutes++;

			printTimeStamp(frame);
			System.out.println("================================================================================");

			System.out.println(String.format(
					"points/minute: %d after %d minutes. Expected result: %d",
					score / gameMinutes, gameMinutes, 5 * score / gameMinutes));
//			System.out.println("avg. game speed: "+ getAverageGameSpeed(frame) +"%");
//			System.out.println("overallThinkTime: " + overallThinkTime + " millisecs in " + frame.getFrameId() + " frames");
			System.out.println(String.format(
					"ms/frame: %.1f, usage: %d%%", 
					(double) overallThinkTime / (double) frame.getFrameId(), 
					(int) ((double) overallThinkTime / (double) frame.getFrameId() * 6d)));
			System.out.println(String.format("rotationByte errors: %d, rate: %.01f%%", 
					badRotationBytes, 
					(double) badRotationBytes / (double) frame.getFrameId() * 100));
			
			// after 5min - finish the game
			if (limitGameTime && gameMinutes > 4) {
				System.out.println("GAME OVER - final score: "+score);
				gameOver = true; 
			}
			
			System.out.println("================================================================================");
		}	
	}

	/**
	 * Decide whether to shoot or not.
	 * 
	 * @param frame
	 */
	private void shoot(FrameInfo frame) {

		// we can't shoot without a plan
		if (!hasShootingPlan()) return;
		
		SpaceShip ship = frame.getSpaceShip();
		Target target = (Target) frame.getObjectByUid(shootingPlan.getTargetUid());

		if (target == null || ship.getRotationByte() == null) return;
		
		if (Math.abs(ship.getRotationByte() - shootingPlan.getRotByte()) > 1) {
//		if (shootingPlan.getTurningTime() > 0) {
			// we didn't reach shooting direction yet
		}
		else if (shootingPlan.getWaitingTime() > 0) {
//			System.out.println("waiting to fire: "+frame.getFrameId()+"/"+shootingPlan.getTimeToShoot());
			// indicate that we're waiting
			ship.setTargetLocked(true);
		}
		else if (frame.getOwnBulletCount() >= 4) {
			// don't even try to shoot if it's impossible
			ship.setCannotShoot(true);
		}
		else if (frames.size() > 0 && frames.getLast().wasPressedFire()) {
			// don't shoot twice within 2 frames
		} 
		else {
			// FIRE!!!
			//System.out.println("FIRE: "+shootingPlan);
			// 
			pressFire();
			shootingPlan.incShotsFired();
			
			if (shootingPlan.getShotsFired() < shootingPlan.getShotsPlanned()) {
				// keep the plan to fire some more shots
//				System.out.println(String.format("SALVE!!! shot no. %d of %d",
//						shootingPlan.getShotsFired(), 
//						shootingPlan.getShotsPlanned()));
				shootingPlan.addWaitingTime(1);
			}
			else {
				// tell the target, it has to die
				target.setUnderFire(frame.getFrameId());
				invalidateShootingPlan(frame);
			}
		}
	}
	
	/**
	 * Turn the ship.
	 * 
	 * @param frame
	 */
	private void turn(FrameInfo frame) {
		
		SpaceShip ship = frame.getSpaceShip();
		if (ship.getRotationByte() == null) { 
			/*
			 *  solange unsere richtung noch nicht justiert ist, erstmal
			 *  nur drehen bis ein sync-winkel vorbeikommt.
			 */ 
			//System.out.println("turning to force sync!");
			keys |= BUTTON_LEFT;
		}
		// zurück in die mitte wenn es keine ziele gibt
		else if (!frame.hasAsteroids() && !frame.hasUfo()) {
			if (ship.getDistance(targetSelector.getCenterTarget()) < HOME_AREA_RADIUS) {
				// nicht drehen - wir sind schon in der mitte und es gibt nix zu tun
			}
			else {
				// in richtung mitte drehen
				double deltaAngle = ship.getDeltaAngleFromViewDirection(
						targetSelector.getCenterTarget());
				if (deltaAngle > ONE_TURN_ANGLE) {
					keys = BUTTON_LEFT;
				}
				else if (deltaAngle < -ONE_TURN_ANGLE) {
					keys = BUTTON_RIGHT;
				}
			}
		}
		else if (!hasShootingPlan()) {
			// kein plan! 
			
			// FIXME turn to the last target even if there is no plan
			// this will save some time at the end of a level
		}
		else {
			// ziel anvisieren
			/*
			 * compare shooting direction with the expected ship rotation for 
			 * the next frame, because this is the direction the shot will have.
			 */
//			int rotation = 
//				Helper.getAssertedRotationByteAtShot(ship, frames.getLast());
//			
//			// accept a little tolerance
//			if (Math.abs(getShootingPlan().getRotByte() - rotation) > 1) {
			if (shootingPlan.getTurningTime() > 0) {
				
				if (shootingPlan.getTurningDirection() == Direction.LEFT) {
					keys |= BUTTON_LEFT;
				}
				else if (shootingPlan.getTurningDirection() == Direction.RIGHT) {
					keys |= BUTTON_RIGHT;
				}
			}
		}
	}

	/**
	 * Accelerates the ship.
	 * 
	 * @param frame
	 */
	private void accelerate(FrameInfo frame) {
		
		SpaceShip ship = frame.getSpaceShip();
		if (ship.getRotationByte() == null) return;
		
		/*
		 * Wann, wie, wohin fliegen?
		 * 1. Gebot: Du sollst nicht fliegen!
		 * 2. nie zu schnell fliegen - wir können nicht bremsen,
		 * 3. versuchen, in die Mitte zu kommen, immer - da die kollisionserkennung nicht über den rand wrappt!
		 * 4. nur beschleunigen, wenn schiff in die zielrichtung zeigt.
		 */
		if (
				// gaaaanz langsam - alles andere geht schief
				ship.getVelocity() < 1.0d &&
				//
				ship.getDistance(targetSelector.getCenterTarget()) > HOME_AREA_RADIUS &&
				// nur beschleunigen, wenn wir zufällig genau in richtung mitte gucken (auflösung 32steps == 11,25°)
				(Math.abs(ship.getExactDirectionAngle() - ship.getDeltaAngle(targetSelector.getCenterTarget())) < 0.2d)
		) {
			keys |= BUTTON_THRUST;
		}
	}

	/**
	 * Make hyperspace jump if being hit is unavoidable.
	 * 
	 * @param frame
	 */
	private void rescue(FrameInfo frame) {
		
		SpaceShip ship = frame.getSpaceShip();

		// kollidierenden Asteroiden suchen
		if (TargetSelector.findCollidingAsteroid(ship, frame) != null) {
			
			pressHyperSpace(frame, "Asteroid");
		}

		// schüsse in der nähe suchen - um evtl. zu flüchten
		if (TargetSelector.findCollidingBullet(ship, frame) != null) {
			
			pressHyperSpace(frame, "Bullet");
		}							
	}
	
	/**
	 * Create shooting solutions for targets that have none.
	 * 
	 * @param currFrame
	 * @param prevFrame
	 */
	private void createMissingShootingSolutions(FrameInfo currFrame, FrameInfo prevFrame) {
		
		if (!currFrame.hasShip()) return;
		
		for (Target t: currFrame.getAllTargets()) {
			
			if (!t.hasShootingSolution() 
				|| t.getShootingSolution().getWaitingTime() < -1) 
			{
				t.setShootingSolution(null);
				Helper.generateShootingSolution(currFrame.getSpaceShip(), t, 
						prevFrame);
			}
		}
	}
	
	/**
	 * Permanently updates target shooting solutions instead of processing all
	 * at once, what could exceed our limited time per frame.
	 * 
	 * @param currFrame The current frame.
	 * @param history List of previous frames. 
	 */
	private void updateShootingSolutions(FrameInfo currFrame) {
		
		FrameInfo prevFrame  = frames.getLast();
		FrameInfo prevFrame2 = frames.get(frames.size() - 2);
		
		for (Target t: currFrame.getAllTargets()) {
			
			if (!t.hasShootingSolution()) continue;
			
			// check for hyperjumps
			if (prevFrame2 != null && prevFrame2.wasPressedHyperspace()) {
				/*
				 * delete all solutions to force creation of new ones
				 * do this 2 frames after the hyperjump-key was sent, to make
				 * use of the frame between - there could still be the chance to 
				 * make a good shot 
				 */
				// System.out.println("deleting all solutions cause of hyperspace");
				t.setShootingSolution(null);
				continue;
			}
			
			ShootingSolution sol = t.getShootingSolution();
			
			// all further checks use the current frame, because pressed keys
			// from the previous frame are already included in solutions
			if (prevFrame.wasPressedLeft() || prevFrame.wasPressedRight()) {
			
				if ((prevFrame.wasPressedLeft() && sol.getTurningDirection() == Direction.LEFT)
				 || (prevFrame.wasPressedRight() && sol.getTurningDirection() == Direction.RIGHT)) {
					// turned into the right direction
					sol.addTurningTime(-1);
				}
				else {
					// turned into the wrong direction
					sol.addTurningTime(+1);
					sol.addWaitingTime(-2);
				}
			}
			else {
				// we missed one step in the plan - decr. waiting time
				t.getShootingSolution().addWaitingTime(-1);
			}
		}
		
		// and now find new solutions for targets that have none
		createMissingShootingSolutions(currFrame, prevFrame);
	}
}
