// ============================================================================
// File:               $File$
//
// Project:            
//
// Purpose:            
//
// Authors:            Harald Boegeholz (original C++ code)
//                     Rammi (port to Java and more)
//
// Copyright Notice:   (c) 2008  Harald Boegeho�z/c't
//                     (c) 2008  Rammi (rammi@caff.de)
//                     This code is in the public domain.
//                     Use at own risk.
//                     No guarantees given.
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.communication;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.io.IOException;
import java.util.*;
import java.util.List;

import de.zappbrannigan.objects.Asteroid;
import de.zappbrannigan.objects.AsteroidPool;
import de.zappbrannigan.objects.Bullet;
import de.zappbrannigan.objects.GameData;
import de.zappbrannigan.objects.GameObject;
import de.zappbrannigan.objects.ShootingSolution;
import de.zappbrannigan.objects.SpaceShip;
import de.zappbrannigan.objects.Target;
import de.zappbrannigan.objects.Ufo;

/**
 *  The info describing one game frame constructed from a received datagram.
 *
 *  This class is part of a solution for a
 *  <a href="http://www.heise.de/ct/creativ/08/02/details/">competition by the German computer magazine c't</a>.
 */
public class FrameInfo
        implements GameData
{
  static final int MAME_DATAGRAM_SIZE  = VECTORRAM_SIZE + 2;
  private static final int ID_BYTE = VECTORRAM_SIZE;
  private static final int PING_BYTE = ID_BYTE+1;

  private static final int MASK_OPCODE = 0xF0;
  private static final int OPCODE_LABS = 0xA0;
  private static final int OPCODE_HALT = 0xB0;
  private static final int OPCODE_JSRL = 0xC0;
  private static final int OPCODE_RTSL = 0xD0;
  private static final int OPCODE_JMPL = 0xE0;
  private static final int OPCODE_SVEC = 0xF0;
  private static final int OPCODE_SHIP = 0x60;

  private static final int SUB_ASTEROID_TYPE1 = 0x8F3;
  private static final int SUB_ASTEROID_TYPE2 = 0x8FF;
  private static final int SUB_ASTEROID_TYPE3 = 0x90D;
  private static final int SUB_ASTEROID_TYPE4 = 0x91A;
  private static final int SUB_UFO            = 0x929;
  private static final int SUB_SHIP			  = 0xA6D;
  private static final int SUB_SPACE		  = 0xB2C;
  private static final int SUB_NUMBER_0		  = 0xADD;
  private static final int SUB_NUMBER_1		  = 0xB2E;
  private static final int SUB_NUMBER_2		  = 0xB32;
  private static final int SUB_NUMBER_3		  = 0xB3A;
  private static final int SUB_NUMBER_4		  = 0xB41;
  private static final int SUB_NUMBER_5		  = 0xB48;
  private static final int SUB_NUMBER_6		  = 0xB4F;
  private static final int SUB_NUMBER_7		  = 0xB56;
  private static final int SUB_NUMBER_8		  = 0xB5B;
  private static final int SUB_NUMBER_9		  = 0xB63;

  /** Frame ID as sent by MAME. */
  private final byte id;
  /** Ping ID as sent by MAME. */
  private final byte ping;
  /** The ufo (if any). */
  private Ufo ufo;
  /** The list of asteroids. */
  private List<Asteroid> asteroids = new LinkedList<Asteroid>();
  /** The spaceship (if any). */
  private SpaceShip spaceShip;
  /** The list of bullets. */
  private List<Bullet> bullets = new LinkedList<Bullet>();
  /** Number of lifes left */
  private int lives;
  /** The time this info was created. */
  private final long receiveTime;
  /** The time used for the last ping (or 0). */
  private final long pingTime;
  /** The game score */
  private int score;
  /** The absolute frame number in this game round. */
  private int frameId;
  /** The key-commands sent in this frame. */
  private byte pressedKeys;
  /** The shooting plan for this frame. */
  private ShootingSolution shootingSolution;
  
  /**
   * Interprete 8 bits as unsigned byte.
   * @param b incomoing byte
   * @return unsigned interpretation
   */
  private static int byteToUnsigned(byte b)
  {
    return b < 0 ? b + 256 : b;
  }

  /**
   *  Constructor.
   *
   *  Extracts frame info from datagram.
   *  @param bytes      datagram bytes
   *  @param pingTimes  time stamps when pings were sent
   *  @throws IOException on i/o errors
   */
  FrameInfo(byte[] bytes, long[] pingTimes, AsteroidPool asteroidPool) throws IOException
  {
    if (bytes.length != MAME_DATAGRAM_SIZE) {
      throw new IOException("Incorrect datagram with size "+bytes.length);
    }
    if ((bytes[1] & MASK_OPCODE) != OPCODE_JMPL) {
      throw new IOException(String.format("Incorrect vector buffer start: %02x%02x", bytes[0], bytes[1]));
    }
    receiveTime = System.currentTimeMillis();
    id = bytes[ID_BYTE];
    ping = bytes[PING_BYTE];
    if (ping != Communication.NO_PING) {
      pingTime = receiveTime - pingTimes[ping];
    }
    else {
      pingTime = 0;
    }
    int vx = 0;
    int vy = 0;
    int vs = 0;
    int v1x = 0;
    int v1y = 0;
    int dx = 0;
    int dy = 0;

    lives = 0;
    score = 0;
    
    // TODO this works only in one player mode
    // refactoring needed to work in two player mode too
    
    boolean possibleShip = false;
    int p = 2;     // skip first two
    while (p < VECTORRAM_SIZE) {
      int highbyte = byteToUnsigned(bytes[p+1]);
      int opcode = highbyte & MASK_OPCODE;
      switch (opcode) {
      case OPCODE_LABS:
        vy = (highbyte & 0x03) << 8 | byteToUnsigned(bytes[p]);
        p += 2;
        highbyte = byteToUnsigned(bytes[p+1]);
        vx = (highbyte & 0x03) << 8 | byteToUnsigned(bytes[p]);
        vs = highbyte >> 4;
        p += 2;
        
        /** initial score position in one player mode */
        if (vx == 100 && vy == 876) {
        	// read score
        	p += 4; // skip idle command
        	for (int i = 4; i >= 0; i--) {

        		int hb = byteToUnsigned(bytes[p+1]);
        		int oc = hb & MASK_OPCODE;
        		
        		if (oc != OPCODE_JSRL) {
        			// error - should never happen
        			break;
        		}
        		
        		int number = 0;
        		switch ((hb & 0x0F) << 8 | byteToUnsigned(bytes[p])) {
	        		case SUB_NUMBER_1: number = 1; break;
	        		case SUB_NUMBER_2: number = 2; break;
	        		case SUB_NUMBER_3: number = 3; break;
	        		case SUB_NUMBER_4: number = 4; break;
	        		case SUB_NUMBER_5: number = 5; break;
	        		case SUB_NUMBER_6: number = 6; break;
	        		case SUB_NUMBER_7: number = 7; break;
	        		case SUB_NUMBER_8: number = 8; break;
	        		case SUB_NUMBER_9: number = 9; break;
        		}
        		score += number * Math.pow(10, i);
        		p += 2;
        	}
        }
        
        break;

      case OPCODE_HALT:
        // p += 2;
        return;

      case OPCODE_JSRL:
        switch ((highbyte & 0x0F) << 8  | byteToUnsigned(bytes[p])) {
        case SUB_ASTEROID_TYPE1:
            asteroids.add(asteroidPool.requestAsteroid(vx * 8, vy * 8, vs, 0));
          break;
        case SUB_ASTEROID_TYPE2:
          asteroids.add(asteroidPool.requestAsteroid(vx * 8, vy * 8, vs, 1));
          break;
        case SUB_ASTEROID_TYPE3:
          asteroids.add(asteroidPool.requestAsteroid(vx * 8, vy * 8, vs, 2));
          break;
        case SUB_ASTEROID_TYPE4:
          asteroids.add(asteroidPool.requestAsteroid(vx * 8, vy * 8, vs, 3));
          break;
        case SUB_UFO:
          ufo = new Ufo(vx * 8, vy * 8, vs);
          break;
        case SUB_SHIP:
          lives++;
          break;
        }
        p += 2;
        break;

      case OPCODE_RTSL:
        // p += 2;
        return;

      case OPCODE_JMPL:
        //p += 2;
        return;

      case OPCODE_SVEC:
        p += 2;
        break;

      default:
        if (spaceShip == null) {
          dy = (highbyte & 0x03) << 8 | byteToUnsigned(bytes[p]);
          if ((highbyte & 0x04) != 0) {
            dy = -dy;
          }
          p += 2;
          highbyte = byteToUnsigned(bytes[p+1]);
          dx = (highbyte & 0x03) << 8 | byteToUnsigned(bytes[p]);
          if ((highbyte & 0x04) != 0) {
            dx = -dx;
          }
          int vz = highbyte >> 4;
          if (dx == 0  &&  dy == 0) {
            if (vz == 15) {
              bullets.add(new Bullet(vx * 8, vy * 8));
            }
          }
          if (dx != 0  &&  dy != 0) {
            if (opcode == OPCODE_SHIP  &&  vz == 12) {
              if (possibleShip) {
                if (spaceShip == null) {
                  spaceShip = new SpaceShip(vx * 8, vy * 8, (v1x - dx) * 8, (v1y - dy) * 8);
                }
                possibleShip = false;
              }
              else {
                v1x = dx;
                v1y = dy;
                possibleShip = true;
              }
            }
          }
          else if (possibleShip) {
            possibleShip = false;
          }
          p += 2;
        }
        else {
          p += 4;
        }
        break;
      }
    }
  }

  /**
   *  Draw all game objects.
   *  @param g graphics context
   */
  public void draw(Graphics2D g) {

	// if game is running, display score and time
	if (frameId > 0) {
		g.setColor(Color.white);
		AffineTransform trafo = new AffineTransform();
		trafo.concatenate(AffineTransform.getScaleInstance(1, -1));
		g.setFont(new Font("Dialog", Font.PLAIN, 160).deriveFont(trafo));
		g.drawString(String.format(
			"%.01fk (%.01fk/min; %.0fk exp.); %dL; %d:%02d",
			(double) score / 1000d, 
			(float) score / (float) frameId * 3.6d,
			(float) score / (float) frameId * 18d,
			lives,
			(int)(frameId/3600),
			(int)(frameId/60)%60),
			100, MAX_Y - 250
		);
	}

    for (Asteroid asteroid: asteroids) {
      asteroid.draw(g);
    }
    
    if (ufo != null) {
      ufo.draw(g);
    }
    
    if (spaceShip != null) {
      spaceShip.draw(g);
    }
    
    for (Bullet bullet: bullets) {
      bullet.draw(g);
    }
    
 	if (shootingSolution != null && shootingSolution.getHitPoint() != null) {
    	g.setColor(Color.red);
    	g.fillOval(
    			shootingSolution.getHitPoint().x - 64, 
    			shootingSolution.getHitPoint().y - 64, 
    			128, 128);
    }
  }

  /**
   *  Get the frame id.
   *
   *  The frame id is a counter which is incremented by mame each time a frame is sent.
   *  @return frame id
   */
  public byte getId()
  {
    return id;
  }

  /**
   *  Get the ping associated with this frame.
   *  @return a number between <code>1</code> and <code>255</code> if there is a ping associated with this frame,
   *          otherwise <code>0</code>
   */
  public int getPing()
  {
    return ping;
  }

  /**
   *  Get the ping time associated with this frame.
   *  @return ping time or <code>0</code> if there is no ping associated with this frame
   */
  public long getPingTime()
  {
    return pingTime;
  }

  /**
   *  Get the time when this frame was received.
   *  @return receive time (compare System.currentTimeMillis())
   */
  public long getReceiveTime()
  {
    return receiveTime;
  }

  /**
   *  Get the ufo.
   *  @return ufo or <code>null</code> if no ufo is present
   */
  public Ufo getUfo()
  {
    return ufo;
  }

  /**
   *  Get the asteroids.
   *
   *  @return the asteroids
   */
  public Collection<Asteroid> getAsteroids()
  {
    return Collections.unmodifiableCollection(asteroids);
  }

  /**
   *  Get the space ship.
   *  @return space ship or <code>null</code> if there is no space ship present
   */
  public SpaceShip getSpaceShip()
  {
    return spaceShip;
  }

  /**
   *  Get the bullets.
   *
   *  @return bullets
   */
  public Collection<Bullet> getBullets()
  {
    return Collections.unmodifiableCollection(bullets);
  }

  /**
   * Get number of lives
   *
   * @return Number of lives
   */
  public int getLives() {
	  
	  return lives;
  }
  
  /**
   * Get the current game score.
   * 
   * @return Score points.
   */
  public int getScore() {
	  return score;
  }
  
  /**
   * Returns the number of own bullets currently on their way.
   * 
   * @return Number of own bullets.
   */
  public int getOwnBulletCount() {
	  int count = 0;
	  
	  for (Bullet b : bullets) {
		  if (b.isOwnShot()) {
			  count++;
		  }
	  }
	  
	  return count;
  }

	/**
	 * @return the frameId
	 */
	public int getFrameId() {
		return frameId;
	}

	/**
	 * @param frameId the frameId to set
	 */
	public void setFrameId(int frameId) {
		this.frameId = frameId;
	}
  
  /**
   * Stores the keys pressed at this frame.
   * 
   * @param keys
   */
  public void setPressedKeys(byte keys) {
	  
	  pressedKeys = keys;
  }
  
  /**
   * Returns the state of button at this frame.
   * 
   * @return True if left was pressed.
   */
  public boolean wasPressedLeft() {
	  
	  return (pressedKeys & BUTTON_LEFT) > 0;
  }
  
  /**
   * Returns the state of button at this frame.
   * 
   * @return True if right was pressed.
   */
  public boolean wasPressedRight() {
	  
	  return (pressedKeys & BUTTON_RIGHT) > 0;
  }
  
  /**
   * Returns the state of button at this frame.
   * 
   * @return True if fire was pressed.
   */
  public boolean wasPressedFire() {
	  
	  return (pressedKeys & BUTTON_FIRE) > 0;
  }
  
  /**
   * Returns the state of button at this frame.
   * 
   * @return True if hyperspace was pressed.
   */
  public boolean wasPressedHyperspace() {
	  
	  return (pressedKeys & BUTTON_HYPERSPACE) > 0;
  }
	
	/**
	 * Get an object from this frame by it's uid.
	 * 
	 * @param uid
	 * @return
	 */
	public GameObject getObjectByUid(int uid) {
		
		if (uid == 0) return null;
		
		GameObject object = null;

		if (hasUfo() && ufo.getUid() == uid) {
			object = ufo;
		}
		
		if (object == null && hasShip()) {
			if (spaceShip.getUid() == uid) object = spaceShip;
		}
		
		if (object == null && hasBullets()) {
			for (Bullet b: bullets) {
				if (b.getUid() == uid) {
					object = b;
					break;
				}
			}
		}
		
		if (object == null && hasAsteroids()) {
			for (Asteroid ast: asteroids) {
				if (ast.getUid() == uid) {
					object = ast;
					break;
				}
			}
		}
		
		return object;
	}
	
	/**
	 * Returns TRUE if an ufo is present.
	 * 
	 * @return True if this frame contains an ufo.
	 */
	public boolean hasUfo() {
		
		return (this.ufo != null);
	}
	
	/**
	 * Returns TRUE if the ship is present.
	 * 
	 * @return True if this frame contains a ship.
	 */
	public boolean hasShip() {
		
		return (this.spaceShip != null);
	}
	
	/**
	 * Returns TRUE if any asteroids are present.
	 * 
	 * @return True if this frame contains asteroids.
	 */
	public boolean hasAsteroids() {
		
		return (this.asteroids.size() > 0);
	}
	
	/**
	 * Returns TRUE if any bullets are present.
	 * 
	 * @return True if this frame contains bullets.
	 */
	public boolean hasBullets() {
		
		return (this.bullets.size() > 0);
	}

	/**
	 * @return the shootingSolution
	 */
	public ShootingSolution getShootingSolution() {
		return shootingSolution;
	}

	/**
	 * @param shootingSolution the shootingSolution to set
	 */
	public void setShootingSolution(ShootingSolution shootingSolution) {
		this.shootingSolution = shootingSolution;
	}

	/**
	 * Returns a list of all targets within this frame.
	 *
	 * @param currFrame
	 * @return
	 */
	public ArrayList<Target> getAllTargets() {
		
		ArrayList<Target> targets = new ArrayList<Target>();
		
		if (hasUfo()) {
			targets.add(getUfo());
		}
		
		if (hasAsteroids()) {
			targets.addAll(getAsteroids());
		}
		return targets;
	}
}