// ============================================================================
// File:               $File$
//
// Project:            Asteroids game solver.
//
// Purpose:            
//
// Author:             Rammi
//
// Copyright Notice:   (c) 2008  Rammi (rammi@caff.de)
//                     This code is in the public domain.
//                     Use at own risk.
//                     No guarantees given.
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.objects;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;

/**
 * Asteroid object in automated player of Asteroids game.
 *
 * This class is part of a solution for a
 * <a href="http://www.heise.de/ct/creativ/08/02/details/">competition by the German computer magazine c't</a>.
 */
public class Asteroid
        extends Target
{
  /** Scaling for shapes when drawing big asteroid. */
  private final static int NORMAL_SCALE = 8 * 8;

  /**
   * Definition of possible asteroid sizes.
   */
  public enum AsteroidSize {
	  /** Asteroid has size: small. */
	  SMALL,
	  /** Asteroid has size: middle. */
	  MIDDLE, 
	  /** Asteroid has size: big. */
	  BIG
  }
  private AsteroidSize asteroidSize;
  /** Scaling of this asteroid. */
  private int scale;
  /** Type of this asteroid. There are 4 different types numbered from 0 to 3. */
  private int type;

  private static final GeneralPath PATH0 = new GeneralPath();
  private static final GeneralPath PATH1 = new GeneralPath();
  private static final GeneralPath PATH2 = new GeneralPath();
  private static final GeneralPath PATH3 = new GeneralPath();
  static {
    PATH0.moveTo(1, 0);
    PATH0.lineTo(4, 1);
    PATH0.lineTo(4, 2);
    PATH0.lineTo(1, 4);
    PATH0.lineTo(-2, 4);
    PATH0.lineTo(-1, 3);
    PATH0.lineTo(-4, 3);
    PATH0.lineTo(-4, -1);
    PATH0.lineTo(-2, -4);
    PATH0.lineTo(1, -3);
    PATH0.lineTo(2, -4);
    PATH0.lineTo(4, -2);
    PATH0.closePath();

    PATH1.moveTo(3, 0);
    PATH1.lineTo(4, 2);
    PATH1.lineTo(2, 4);
    PATH1.lineTo(0, 2);
    PATH1.lineTo(-2, 4);
    PATH1.lineTo(-4, 2);
    PATH1.lineTo(-4, -2);
    PATH1.lineTo(-2, -4);
    PATH1.lineTo(1, -4);
    PATH1.lineTo(4, -2);
    PATH1.closePath();

    PATH2.moveTo(4, 1);
    PATH2.lineTo(2, 4);
    PATH2.lineTo(-1, 4);
    PATH2.lineTo(-4, 1);
    PATH2.lineTo(-2, 0);
    PATH2.lineTo(-4, -1);
    PATH2.lineTo(-2, -4);
    PATH2.lineTo(0, -1);
    PATH2.lineTo(0, -4);
    PATH2.lineTo(2, -4);
    PATH2.lineTo(4, -1);
    PATH2.closePath();

    PATH3.moveTo(2, 1);
    PATH3.lineTo(4, 2);
    PATH3.lineTo(2, 4);
    PATH3.lineTo(0, 3);
    PATH3.lineTo(-2, 4);
    PATH3.lineTo(-4, 2);
    PATH3.lineTo(-3, 0);
    PATH3.lineTo(-4, -2);
    PATH3.lineTo(-2, -4);
    PATH3.lineTo(-1, -3);
    PATH3.lineTo(2, -4);
    PATH3.lineTo(4, -1);
    PATH3.closePath();
  }
  /** The asteroid shapes sorted by type. */
  private static final Shape[] SHAPES = new Shape[] {
          PATH1,
          PATH3,
          PATH2,
          PATH0,
  };

  /**
   * Constructor.
   * @param x position (x coordinate)
   * @param y position (y coordinate)
   * @param size size (as given by game, so 0 is big and 14 is small)
   * @param type type of asteroid (0, 1, 2, or 3)
   */
  public Asteroid(int x, int y, int size, int type)
  {
	  super(x, y);
	  initializeVariables(x, y, size, type);
  }

  /**
   * Initializes all variables of an Asteroid 
   * 
   * @param x Position x
   * @param y Position y
   * @param size Size
   * @param type Type
   */
  public void initializeVariables(int x, int y, int size, int type) {
	  
	setLocation(x, y);
	setSize(size);
	setType(type);
	setVelocity(0, 0);
	setAge(0);
	currentTarget = false;
	critical = false;
	underFire = false;
	shotTime = -1;
	vScale = 16;
	uid = -1;
	shootingSolution = null;
	trackingComplete = false;
  }
  
  /**
   * Set size of asteroid
   * 
   * @param size
   */
  public void setSize(int size) {

	  switch (size) {
	    case 0:
	      this.scale = NORMAL_SCALE;
	      asteroidSize = AsteroidSize.BIG;
	      break;

	    case 15:
	      this.scale = NORMAL_SCALE/2;
	      asteroidSize = AsteroidSize.MIDDLE;
	      break;

	    case 14:
	      this.scale = NORMAL_SCALE/4;
	      asteroidSize = AsteroidSize.SMALL;
	      break;
	    }
  }
  
  /**
   * Sets the type of the Asteroid
   * 
   * @param type Type
   */
  public void setType(int type) {
	  
	    assert(type >= 0  &&  type < SHAPES.length);
	    this.type = type;
  }
  
  /**
   *  Get the size of the object.
   *
   *  The size returned by this method is half the length of a square which contains the object,
   *  so the object's bounding box is between (x - size, y - size) and (x + size, y + size).
   *  @return object size
   */
  @Override
  public int getSize()
  {
    return 4*scale;
  }

  /**
   * Get the type of the asteroid
   * 
   * @return Type
   */
  @Override
  public int getType()
  {
    return type;
  }

  /**
   *  Draw the object
   *  
   *  @param g graphics context
   */
  @Override
  public void draw(Graphics2D g)
  {
    g.setColor(getColor());
    AffineTransform trafo = new AffineTransform(scale, 0, 0, scale, x, y);
    if (currentTarget) {
        g.fill(trafo.createTransformedShape(SHAPES[type]));
    }
    else {
        g.draw(trafo.createTransformedShape(SHAPES[type]));
    }

	super.draw(g);
  }

  /**
   * Returns a string representation of the object.
   *
   * @return a string representation of the object.
   */
  @Override
  public String toString()
  {
    return String.format("Asteroid@(%d,%d)*%d?%d, vx:vy=%d:%d", x, y, getSize(), type, vx,vy);
  }

  /**
   * @see Target#getPoints()
   */
	@Override
	public int getPoints() {
		
		int result = 0;
	
	    switch (asteroidSize) {
	    case BIG:
	    	result = 20;
	    	break;
	    case MIDDLE:
	    	result = 50;
	    	break;
	    case SMALL:
	    	result = 100;
	    	break;
	    }
	
		return result;
	}

	/**
	 * Get the worth of this asteroid's debris.
	 * 
	 * @return The worth of this asteroid's debris.
	 */
	public int getIncludedPoints() {
		
		int result = 0;
		
		if (asteroidSize == AsteroidSize.BIG) {
			
			result += 2 * 50 + 4 * 100;
		}
		else if (asteroidSize == AsteroidSize.MIDDLE) {
			
			result += 2 * 50;
		}
		
		return result;
	}
	
	/**
	 * @see GameObject#getVelocityColor()
	 */
	@Override
	public Color getVelocityColor() {
		return Color.yellow;
	}
	
//	/**
//	 * @see Object#equals(Object)
//	 */
//	@Override
//	public boolean equals(Object a) {
//		
//		boolean result = super.equals(a);
//
//		if (a instanceof Asteroid) {
//			
//			Asteroid asteroid = (Asteroid) a;
//			
//			if (getSize() != asteroid.getSize()) {
//				return false;
//			}
//		
//			if (getType() != asteroid.getType()) {
//				return false;
//			}
//		}
//		else {
//			result = false;
//		}
//		
//		return result;
//	}
	
	/**
	 * Get the asteroid size.
	 * 
	 * @return The asteroid size.
	 */
	public AsteroidSize getNativeSize() {
		
		return this.asteroidSize;
	}
}