// ============================================================================
// File:               $File$
//
// Project:            
//
// Purpose:            
//
// Author:             Rammi
//
// Copyright Notice:   (c) 2008  Rammi (rammi@caff.de)
//                     This code is in the public domain.
//                     Use at own risk.
//                     No guarantees given.
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.objects;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.util.LinkedList;

import de.zappbrannigan.ai.Helper;
import de.zappbrannigan.communication.FrameInfo;

/**
 * Bullet object in automated player of Asteroids game.
 * 
 * Bullets are either shot by the ship or by an ufo.
 * 
 * This class is part of a solution for a <a
 * href="http://www.heise.de/ct/creativ/08/02/details/">competition by the
 * German computer magazine c't</a>.
 */
public class Bullet extends GameObject {

	/** True if this bullet was fired by our own. */
	private boolean isOwnShot;
	/** The point this bullet has been seen for the first time. */
	private Point origin;
	/** The shooting solution this bullet should follow. */
	private ShootingSolution shootingSolution;
	
	/**
	 * Constructor.
	 * 
	 * @param x
	 *            x coordinate
	 * @param y
	 *            y coordinate
	 */
	public Bullet(int x, int y) {
		super(x, y);
	}

	/**
	 * Get the size of the object.
	 * 
	 * The size returned by this method is half the length of a square which
	 * contains the object, so the object's bounding box is between (x - size, y -
	 * size) and (x + size, y + size).
	 * 
	 * @return object size
	 */
	@Override
	public int getSize() {
		return 1;
	}

	/**
	 * Marks a bullet as critical
	 * 
	 * @param b
	 *            TRUE if the bullet is critical
	 */
	public void makeApproaching(boolean b) {

		critical = b;
	}

	/**
	 * @see GameObject#getVelocityColor()
	 */
	protected Color getVelocityColor() {
	
		if (critical)
			return Color.red;
		else if (isOwnShot) 
			return Color.blue;
		else
			return new Color(125, 0, 0);
	}

	/**
	 * Draw the object.
	 * 
	 * @param g
	 *            graphics context
	 */
	@Override
	public void draw(Graphics2D g) {
		
// DEBUG output to determine real torpedo speeds		
//System.out.println(String.format("Bullet#%d: Angle=%d, RotByte=%d, vx=%d, vy=%d, v=%d",
//		this.age,
//		(int) Math.round(Math.toDegrees(getVelocityAngle())+360) % 360,
//		(int) Math.round(Math.toDegrees(Helper.getAngleFromRotByte(rotationByte))+360) % 360,
//		(int) Math.round(vx),
//		(int) Math.round(vy),
//		(int) Math.round(getVelocity())));

		drawVelocityVector(g, getVelocityColor());
		g.setColor(Color.white);
		g.fillOval(x - 24, y - 24, 50, 50);
		
		if (shootingSolution != null) {
			drawAimingPoint(g);
		}
	}

	/**
	 * Overwrites drawVelocity, arrow heads end at bullet
	 * 
	 * @param g
	 * @param color
	 */
	@Override
	protected void drawVelocityVector(Graphics2D g, Color color) {
		if (hasKnownVelocity()) {
			g.setColor(color);			
			g.drawLine(x, y, x - (int) (vScale * vx), y - (int) (vScale * vy));
			AffineTransform at = AffineTransform.getTranslateInstance(x, y);
			at.concatenate(AffineTransform.getRotateInstance(getVelocityAngle()));
			g.fill(at.createTransformedShape(ARROW_HEAD));
		}
	}

	/**
	 * Draws the asserted bullet-target-hitpoint.
	 * 
	 * @param g
	 */
	private void drawAimingPoint(Graphics2D g) {

		if (origin != null) {
			g.setColor(new Color(24, 24, 24));
			g.drawLine(
				origin.x, 
				origin.y, 
				origin.x + (int)Math.round(Math.cos(Helper.getAngleFromRotByte(shootingSolution.getRotByte())) * EXTENT),
				origin.y + (int)Math.round(Math.sin(Helper.getAngleFromRotByte(shootingSolution.getRotByte())) * EXTENT));
		}
		
		int radius = 60;
		int centerX = shootingSolution.getHitPoint().x;
		int centerY = shootingSolution.getHitPoint().y;
		g.setColor(new Color(128, 0, 0));
		g.drawOval(shootingSolution.getHitPoint().x - radius, 
				shootingSolution.getHitPoint().y - radius, radius*2, radius*2);
		g.drawLine(centerX + radius/2, centerY, centerX + radius + radius/2, centerY);
		g.drawLine(centerX - radius/2, centerY, centerX - radius - radius/2, centerY);
		g.drawLine(centerX, centerY + radius/2, centerX, centerY + radius + radius/2);
		g.drawLine(centerX, centerY - radius/2, centerX, centerY - radius - radius/2);
	}
	
	/**
	 * Returns the position this shot comes from.
	 * 
	 * @return The point of shooting.
	 */
	public Point getOrigin() {
		return origin;
	}

	/**
	 * Sets the postion this bullet is shot from.
	 * Useful to distinguish between own and ufo bullets. 
	 * 
	 * @param p
	 */
	public void setOrigin(Point p) {
		origin = p;
	}
	
	/**
	 * Sets the own-shot flag.
	 * 
	 * @param b
	 */
	public void markOwnShot(boolean b) {
		isOwnShot = b;
	}
	
	/**
	 * Returns true if this shot has been fired by our own.
	 * 
	 * @return True if this shot comes from ourself.
	 */
	public boolean isOwnShot() {
		return isOwnShot;
	}
	
	/**
     * Merges all necessary information from the predecessor into the current object.  
	 * 
	 * @param b The predecessing bullet.
	 * @param history A list of historic frames. 
	 */
	public void merge(Bullet b, LinkedList<FrameInfo> history) {

		// merge common properties (uid, age, velocity)
		super.merge(b, history);

		// handle new bullets
		if (age == 1) {

			/*
			 * just some theory explained:
			 * 	t-3 the fire-key was pressed - this frame contains the shooting 
			 * 		solution for the bullet
			 * 	t-2 the ship has the direction that the bullet will inherit
			 * 	t-1 the bullet appears
			 * 	t   the bullet has made one step, we can now astimate it's direction
			 */
			
			// the frame when the fire button was pressed - this determines the target
			FrameInfo frameShotAt = history.get(history.size() - 4);
			// the frame when this bullet was ignited - this determines the direction
			FrameInfo prevFrame = history.get(history.size() - 3); 

			if (frameShotAt.getShootingSolution() != null) {
				setShootingSolution(frameShotAt.getShootingSolution());
			}
			
			// identify and mark own torpedos
			SpaceShip ship = prevFrame.getSpaceShip();
	        if (ship != null && getDistance(ship) < 250) {
				setOrigin(ship.getCenter());
	      	  	markOwnShot(true);
	        } 
		}
		// handle older bullets
		else {
			
			setOrigin(b.getOrigin());
			setShootingSolution(b.getShootingSolution());
      	  	markOwnShot(b.isOwnShot());
        }
	}

	/**
	 * @return the aimingPoint
	 */
	public Point getAimingPoint() {
		return shootingSolution.getHitPoint();
	}

	/**
	 * @return the rotationByte
	 */
	public int getRotationByte() {
		return shootingSolution.getRotByte();
	}

	/**
	 * @return the targetUid
	 */
	public int getTargetUid() {

		int targetUid = 0;
		
		if (shootingSolution != null) {
			targetUid = shootingSolution.getTargetUid(); 
		}
		
		return targetUid;
	}

	/**
	 * @return the shootingSolution
	 */
	public ShootingSolution getShootingSolution() {
		return shootingSolution;
	}

	/**
	 * @param shootingSolution the shootingSolution to set
	 */
	public void setShootingSolution(ShootingSolution shootingSolution) {
		this.shootingSolution = shootingSolution;
	}
}
