// ============================================================================
// File:               $File$
//
// Project:            
//
// Purpose:            
//
// Author:             Rammi
//
// Copyright Notice:   (c) 2008  Rammi (rammi@caff.de)
//                     This code is in the public domain.
//                     Use at own risk.
//                     No guarantees given.
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.objects;

import java.awt.*;
import java.awt.geom.GeneralPath;
import java.awt.geom.AffineTransform;
import java.util.LinkedList;

import de.zappbrannigan.ai.Helper;
import de.zappbrannigan.communication.FrameInfo;

/**
 * Generic game object in Asteroid game.
 * 
 * A game object has a position and a size and it can be drawn.
 * 
 * This is part of a solution for a competition <a
 * href="http://www.heise.de/ct/creativ/08/02/details/">by the German computer
 * magazine c't</a>
 */
public abstract class GameObject 
		implements GameData {

	/** Size of head of velocity arrow. */
	private static final int ARROW_HEAD_SIZE = 8 * 8;

	/** The arrow head for velocity drawings. */
	protected static final GeneralPath ARROW_HEAD = new GeneralPath();
	static {
		ARROW_HEAD.moveTo(0, 0);
		ARROW_HEAD.lineTo(-ARROW_HEAD_SIZE, ARROW_HEAD_SIZE);
		ARROW_HEAD.lineTo(-ARROW_HEAD_SIZE, -ARROW_HEAD_SIZE);
		ARROW_HEAD.closePath();
	}

	/** The x coordinate. 0..8191 */
	protected int x;
	/** The y coordinate. 1024..7167 (y is point upwards) */
	protected int y;

	/** The x coordinate of the velocity vector. */
	protected int vx;
	/** The y coordinate of the velocity vector. */
	protected int vy;
	/** Age of this object since first recognition (in frames) */
	protected int age = 0;
	/** The unique id of this object to identify it over multiple frames. */
	protected int uid;
	/** Holds the information that an object is critical */
	protected boolean critical = false;
	/** Is tracking of this object complete. */
	protected boolean trackingComplete = false;
	/** magnification factor of velocity vector */
	protected int vScale = 16;
	/** Normal color for game objects */
	protected Color normalColor = Color.gray;

	/**
	 * Constructor.
	 * 
	 * @param x x coordinate
	 * @param y y coordinate
	 */
	protected GameObject(int x, int y) {
		this.x = x;
		this.y = y;
	}

	/**
	 * Get the x coordinate of the objects center.
	 * 
	 * @return center x
	 */
	public int getX() {
		return x;
	}

	/**
	 * Get the y coordinate of the objects center.
	 * 
	 * @return center y
	 */
	public int getY() {
		return y;
	}

	/**
	 * Get the center of the object.
	 * 
	 * @return center point
	 */
	public Point getCenter() {
		return new Point(x, y);
	}

	/**
	 * Get the size of the object.
	 * 
	 * The size returned by this method is half the length of a square which
	 * contains the object, so the object's bounding box is between (x - size, y -
	 * size) and (x + size, y + size).
	 * 
	 * @return object size
	 */
	public abstract int getSize();

	/**
	 * Get the color for the velocity arrow
	 * 
	 * @return Color for the velocity arrow
	 */
	protected abstract Color getVelocityColor();

	/**
	 * Draw the object.
	 * 
	 * @param g graphics context
	 */
	public void draw(Graphics2D g) {
		
		drawVelocityVector(g, getVelocityColor());
	}

	/**
	 * Get the shortest vector from this to another game object.
	 * 
	 * Because of the torus topology it may be shorter going across the borders.
	 * 
	 * @param obj other object
	 * @return shortest vector from this to the other object
	 */
	public Point getDelta(GameObject obj) {

		return getDelta(obj.getCenter());
	}

	/**
	 * Get the shortest vector from this to a point.
	 * 
	 * @param tx
	 * @param ty
	 * @return shortest vector from this to a point
	 */
	public Point getDelta(int tx, int ty) {

		return getDelta(new Point(tx, ty));
	}

	/**
	 * Get the shortest vector from this to a point.
	 * 
	 * @param point
	 * @return shortest vector from this to a point
	 */
	public Point getDelta(Point point) {

		return Helper.getDelta(getCenter(), point);
	}

	/**
	 * Returns the angle from this to another object.
	 * 
	 * @param o The other object.
	 * @return The angle in radians.
	 */
	public double getDeltaAngle(GameObject o) {

		return getDeltaAngle(o.getCenter());
	}

	/**
	 * 
	 * @param p The other point.
	 * @return The angle in radians.
	 */
	public double getDeltaAngle(Point p) {

		Point delta = getDelta(p);
		return Helper.normalizeAngle(Math.atan2(delta.y, delta.x));
	}

	/**
	 * Get the shortest distance from this to another game object.
	 * 
	 * @param obj other object
	 * @return shortest distance from this to the other object
	 */
	public int getDistance(GameObject obj) {

		return getDistance(obj.getCenter().x, obj.getCenter().y);
	}

	/**
	 * Calculates the distance from the game object to a Point
	 * 
	 * @param p Point
	 * @return Distance to the Point
	 */
	public int getDistance(Point p) {

		return getDistance(p.x, p.y);
	}

	/**
	 * Calculates the distance from the game object to a point
	 * 
	 * @param tx Position x of the point
	 * @param ty Position y of the point
	 * @return Distance to the point
	 */
	public int getDistance(int tx, int ty) {

		Point dist = getDelta(tx, ty);
		return (int) Math.hypot(dist.x, dist.y);
	}

	/**
	 * Set the velocity.
	 *
	 * The velocity is the step between frames.
	 * 
	 * @param x x coordinate of velocity
	 * @param y y coordinate of velocity
	 */
	public void setVelocity(int x, int y) {
		vx = x;
		vy = y;
	}

	/**
	 * Set the location
	 * 
	 * @param x
	 * @param y
	 */
	public void setLocation(int x, int y) {

		this.x = x;
		this.y = y;
	}

	/**
	 * Adjust the velocity assuming that the given object is at the same place
	 * as this object in the last frame.
	 * 
	 * @param o Precessor (<code>null</code> allowed, but then nothing happens)
	 * @param history A list of historic frames.
	 */
	public void adjustVelocityFromDelta(GameObject o,
			LinkedList<FrameInfo> history) {

		// this should never happen!
		if (o == null) return;
		// keep old velocity values - except for the ship!
		if (o.isTrackingComplete() && !(this instanceof SpaceShip)) {
			// if object is perfectly tracked, there is no need to calculate 
			// the velocity again
			setVelocity(o.getVelocityX(), o.getVelocityY());
			setTrackingComplete(true);
		}
		else if (o.getAge() == 0) {
			// the precessor is a new object with unknown velocity values
			Point delta = o.getDelta(this);
			setVelocity(delta.x, delta.y);
		} 
		else {
			// start search 9 frames ago to base calculation on 8 moves
			for (int i = 8; i > 0; i--) {
				// try again if history is too short
				if (history.size() < i + 1) continue;

				// try again if object is not present
				GameObject prevObj = history.get(history.size() - 1 - i).getObjectByUid(uid);
				if (prevObj == null) continue;

				Point delta = prevObj.getDelta(this);
				int newVX = (int) Math.round(delta.x / i);
				int newVY = (int) Math.round(delta.y / i);

				setVelocity(newVX, newVY);
				
				if (i == 8) {
					// after 8 steps we have the exact velocity 
					setTrackingComplete(true);
//					System.out.println("tracking completed for "+this.getUid());
				}
				break;
			}
		}
	}

	/**
	 * Get the velocity in x.
	 * 
	 * @return x component of velocity vector
	 */
	public int getVelocityX() {
		return vx;
	}

	/**
	 * Get the velocity in y.
	 * 
	 * @return y component of velocity vector
	 */
	public int getVelocityY() {
		return vy;
	}

	/**
	 * Get the velocity as hypothenusis of the velocity vector.
	 * 
	 * @return velocity
	 */
	public double getVelocity() {
		return Math.hypot(vx, vy);
	}

	/**
	 * Get the velocity angle.
	 * 
	 * The angle is measured counterclockwise, with <code>0</code> pointing to
	 * the right
	 * 
	 * @return velocity in radians, or -Math.PI if ship has no velocity
	 */
	public double getVelocityAngle() {
		return Math.atan2(vy, vx);
	}

	/**
	 * Get the velocity as a vector. 
	 * 
	 * @return velocity vector
	 */
	public Point getVelocityVector() {
		return new Point(vx, vy);
	}
	
	/**
	 * Has this object a velocity.
	 * 
	 * @return velocity
	 */
	public boolean hasKnownVelocity() {
		return vx != 0 || vy != 0;
	}

	/**
	 * Returns the minimum distance a fly-by object will have if it follows it's
	 * current path.
	 * 
	 * @param obj
	 * @return minimum distance while flying by [px]
	 */
	public int getFlyByDistance(GameObject obj) {
		Point p = obj.getCenter();
		double a = obj.getVelocityY();
		double b = -obj.getVelocityX();
		double c = -(a * p.x + b * p.y);
		double d = (a * x + b * y + c) / Math.hypot(a, b);

		return (int) Math.round(Math.abs(d));
	}

	/**
	 * Draw a vector displaying the current velocity.
	 * 
	 * @param g
	 *            Graphics context
	 * @param color
	 *            Color
	 */
	protected void drawVelocityVector(Graphics2D g, Color color) {
		if (hasKnownVelocity()) {
			g.setColor(color);
			g.drawLine(x, y, x + vScale * vx, y + vScale * vy);
			AffineTransform at = AffineTransform.getTranslateInstance(x
					+ vScale * vx, y + vScale * vy);
			at.concatenate(AffineTransform
					.getRotateInstance(getVelocityAngle()));
			g.fill(at.createTransformedShape(ARROW_HEAD));
		}
	}

	/**
	 * Get the number of frames, we could follow this object's move.
	 * 
	 * @return The number of frames.
	 */
	public int getAge() {

		return age;
	}

	/**
	 * Returns a string representation of the object.
	 * 
	 * @return a string representation of the object.
	 */
	@Override
	public String toString() {
		return String.format("%s@(%d,%d)=>(%d,%d) age: %d", 
				this.getClass(), x, y, vx, vy, age);
	}

	/**
	 * Merges all necessary information from the predecessor into the current
	 * object.
	 * 
	 * @param o
	 *            The predecessor.
	 * @param history
	 *            A list of historic frames.
	 */
	public void merge(GameObject o, LinkedList<FrameInfo> history) {

		if (o != null) {

			if (o.getAge() == 0) {

				// initialize a new object
				uid = AsteroidPool.getAsteroidPool().getNextObjectUid();
			} else {

				// merge data from a previous object
				uid = o.getUid();
			}

			// adjust velocity values
			adjustVelocityFromDelta(o, history);

			// increase object age - means we monitor this object for <age>
			// frames.
			age = o.getAge() + 1;
		}
	}

	/**
	 * Returns the position of this object <code>time</code> frames in the
	 * future.
	 * 
	 * @param time
	 *            The number of frames you'll look ahead.
	 * @return The assumed position.
	 */
	public Point getFuturePosition(int time) {
		Point futurePosition = new Point(x + time * vx, y + time * vy);

		return Helper.normalizePoint(futurePosition);
	}

	/**
	 * Set age of game object
	 * 
	 * @param age
	 */
	public void setAge(int age) {

		this.age = age;
	}

	/**
	 * @return the uid
	 */
	public int getUid() {

		return uid;
	}

	/**
	 * Set the objects unique id.
	 * 
	 * @param uid
	 */
	public void setUid(int uid) {

		this.uid = uid;
	}

	/**
	 * @return the trackingComplete
	 */
	public boolean isTrackingComplete() {
		return trackingComplete;
	}

	/**
	 * @param trackingComplete the trackingComplete to set
	 */
	public void setTrackingComplete(boolean trackingComplete) {
		this.trackingComplete = trackingComplete;
	}

	/**
	 * Returns TRUE if the point is ahead of a moving game object.
	 * 
	 * @param point
	 * @return True if the point is ahead.
	 */
	public boolean isPointAhead(Point point) {
		
		if (point == null) return false;
	
		boolean isAhead = false;
		
		// we need an unwrapped delta to determine the position of the point
		int deltaX = point.x - x;
		int deltaY = point.y - y;
		// get direction to the point - normalized into 4 quadrants 
		int dirObjectX = (deltaX != 0) ? deltaX / Math.abs(deltaX) : 0;
		int dirObjectY = (deltaY != 0) ? deltaY / Math.abs(deltaY) : 0;
		// get direction of the object - normalized into 4 quadrants
		int velObjectX = (vx != 0) ? vx / Math.abs(vx) : 0;
		int velObjectY = (vy != 0) ? vy / Math.abs(vy) : 0;
		// 
		if (dirObjectX == velObjectX && dirObjectY == velObjectY) {
			isAhead = true;
		}
		
		return isAhead;
	}
}
