// ============================================================================
// File:               $File$
//
// Project:            
//
// Purpose:            
//
// Author:             Rammi
//
// Copyright Notice:   (c) 2008  Rammi (rammi@caff.de)
//                     This code is in the public domain.
//                     Use at own risk.
//                     No guarantees given.
//
// Latest change:      $Date$
//
// History:	       $Log$
//=============================================================================
package de.zappbrannigan.objects;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.util.LinkedList;

import de.zappbrannigan.ai.Helper;
import de.zappbrannigan.communication.FrameInfo;

/**
 * Space ship in automated player of Asteroid game.
 * 
 * This is part of a solution for a competition 
 * <a href="http://www.heise.de/ct/creativ/08/02/details/">by the German computer magazine c't</a>
 */
public class SpaceShip 
		extends GameObject 
{

	/** The ship shape for drawing. */
	private static final GeneralPath SHIP_SHAPE = new GeneralPath();

	private static final int SHIP_SIZE = 16 * 8;

	static {
		// defined for 0 angle
		SHIP_SHAPE.moveTo(SHIP_SIZE, 0);
		SHIP_SHAPE.lineTo(-SHIP_SIZE, SHIP_SIZE / 2);
		SHIP_SHAPE.lineTo(-SHIP_SIZE / 2, 0);
		SHIP_SHAPE.lineTo(-SHIP_SIZE, -SHIP_SIZE / 2);
		SHIP_SHAPE.closePath();
	}

	/** The direction of the ship, x coordinate. */
	private final int dx;

	/** The direction of the ship, y coordinate. */
	private final int dy;

	private boolean targetLocked = false;
	
	private boolean cannotShoot = false;
	/** The rotation value at the current frame. */
	private Integer rotationByte = null;

	/**
	 * Constructor.
	 * 
	 * @param x x coordinate of ship center
	 * @param y y coordinate of ship center
	 * @param dx direction x
	 * @param dy direction y
	 */
	public SpaceShip(int x, int y, int dx, int dy) 
	{
		super(x, y);
		this.dx = dx;
		this.dy = dy;
		normalColor = Color.white;
	}

	/**
	 * Get the size of the object.
	 * 
	 * The size returned by this method is half the length of a square which
	 * contains the object, so the object's bounding box is between (x - size, y -
	 * size) and (x + size, y + size).
	 * 
	 * @return object size
	 */
	@Override
	public int getSize() 
	{	
		return SHIP_SIZE;
	}

	/**
	 * Get the x coordinate of the ship's direction.
	 * 
	 * @return ship direction, x coordinate
	 */
	public int getDirX() 
	{
		return dx;
	}

	/**
	 * Get the y coordinate of the ship's direction.
	 * 
	 * @return ship direction, y coordinate
	 */
	public int getDirY() 
	{
		return dy;
	}

	/**
	 * Returns the estimated angle of the ship from is visible representation.
	 * 
	 * The angle is measured counterclockwise, with <code>0</code> pointing to
	 * the right
	 * 
	 * @return viewing angle [-PI..0..PI]
	 */
	public double getDirectionAngle() 
	{
		return Math.atan2(dy, dx);
	}
	
	/**
	 * Returns the exact angle of the ship calculated from the rotation byte.
	 * 
	 * @return Angle in radians
	 */
	public double getExactDirectionAngle() {
		
		return Helper.getAngleFromRotByte(rotationByte);
	}

	/**
	 * Get the objects direction relative to the ship.
	 * 
	 * -1 object aims on you +1 object flys away
	 * 
	 * @param t0
	 * @return objects direction [-1..+1]
	 */
	public double getTargetDirection(GameObject t0) {
		
		// aktueller abstand zum objekt
		int d0 = getDistance(t0);

		// erwarteter abstand im nächsten frame
		double d1 = Math.hypot(
				t0.getX() + t0.getVelocityX() - x, 
				t0.getY() + t0.getVelocityY() - y
		);

		/*
		 * delta distance 
		 * <0 objekt kommt näher, 
		 * >0 objekt entfernt sich
		 */
		double dd = d1 - d0;

		return dd / t0.getVelocity();
	}

	/**
	 * Returns the angle to an object according to the current view direction
	 * 
	 * @param object The object
	 * @return The angle to the object
	 */
	public double getDeltaAngleFromViewDirection(GameObject object) {

		return getDeltaAngle(object) - getExactDirectionAngle();
	}

	/**
	 * Get the color to paint the ship with.
	 * 
	 * @return The ship color.
	 */
	private Color getColor() {

		if (cannotShoot) {
			return Color.red;
		} 
		else if (targetLocked) {
			return Color.green;
		} 
		else {
			return normalColor;
		}
	}

	/**
	 * Draw the object.
	 * 
	 * @param g graphics context
	 */
	@Override
	public void draw(Graphics2D g) 
	{
		// drawQuadrants(g);

		g.setColor(getColor());
		AffineTransform at = AffineTransform.getTranslateInstance(x, y);
		at.concatenate(AffineTransform.getRotateInstance(getDirectionAngle()));
		g.fill(at.createTransformedShape(SHIP_SHAPE));
	}

	/**
	 * Returns a string representation of the object.
	 * 
	 * @return a string representation of the object.
	 */
	@Override
	public String toString() 
	{
		return String.format("SpaceShip@(%d,%d)>(%.1f°)", x, y, 180 * getDirectionAngle() / Math.PI);
	}
	
	/**
	 * Set the target-locked status.
	 * 
	 * @param b True if a target is locked.
	 */
	public void setTargetLocked(boolean b) {
		targetLocked = b;
	}

	/**
	 * Get the target-locked status.
	 * 
	 * @return True if a target is locked.
	 */
	public boolean isTargetLocked() {
		return targetLocked;
	}

	/**
	 * Set to true if the ship can't shoot at this moment.
	 * 
	 * @param b True if the ship can't shoot.
	 */
	public void setCannotShoot(boolean b) {

		cannotShoot = b;
	}

	/**
	 * @return the rotationByte
	 */
	public Integer getRotationByte() {
		
		return rotationByte;
	}

	/**
	 * @param rotationByte the rotationByte to set
	 */
	public void setRotationByte(Integer rotationByte) {
		
		this.rotationByte = rotationByte;
	}

	/**
	 * @see GameObject#getVelocityColor()
	 */
	@Override
	protected Color getVelocityColor() {
		return Color.white;
	}
	
	/**
	 * @see Target#merge(Target, LinkedList)
	 * 
	 * @param ship
	 * @param history
	 */
	public void merge(SpaceShip ship, LinkedList<FrameInfo> history) {
		
		super.merge(ship, history);
		
		// merge rotation byte from previous ship to avoid rotbyte failures
		setRotationByte(ship.getRotationByte());
	}

	/**
	 * Returns true if the point is ahead of the ship at the given rotation.
	 * 
	 * @param point The point to verify.
	 * @param rotation The ships rotation.
	 * @return True if the point is ahead of the ship.
	 */
	public boolean isPointAhead(Point point, int rotation) {
		
		// eliminate points in our back
		if (point == null) return false;

		boolean isAhead = false;
		
		// we need an unwrapped delta to determine the position of the point
		int deltaX = point.x - x;
		int deltaY = point.y - y;
		// get direction of the ship - normalized into 4 quadrants 
		int dirTargetX = (deltaX != 0) ? deltaX / Math.abs(deltaX) : 0;
		int dirTargetY = (deltaY != 0) ? deltaY / Math.abs(deltaY) : 0;
		// get the position of the point normalized into 4 quadrants
		Point dirShip = Helper.predictTorpedoVelocity(this, rotation);
		int dirShipX = (dirShip.x != 0) ? dirShip.x / Math.abs(dirShip.x) : 0;
		int dirShipY = (dirShip.y != 0) ? dirShip.y / Math.abs(dirShip.y) : 0;
		
		if (dirTargetX == dirShipX && dirTargetY == dirShipY) {
			// point is ahead
			isAhead = true;
		}
		
		return isAhead;
	}
}
