package de.zappbrannigan.objects;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.util.LinkedList;
//import java.util.SortedMap;

import de.zappbrannigan.communication.FrameInfo;

/**
 * A possible target in the asteroids game. Can be of type ufo or asteroid. 
 * 
 */
public abstract class Target extends GameObject {

	/** True if this target is the next to be killed. */
	protected boolean currentTarget = false;
	/** True if any torpedos are on the way to kill this target. */
	protected boolean underFire = false;
	/** The frame number when the last shot on this target was fired. */
	protected int shotTime = -1;
	/** A possible shooting solution to destroy this target. */
	protected ShootingSolution shootingSolution = null;
	
	/**
	 * Constructor
	 * 
	 * @param x The initial x position.
	 * @param y The current y position.
	 */
	public Target(int x, int y) {
		super(x, y);
	}
	
	/**
	 * Get the value in points of this object when shot.
	 * 
	 * @return The value in points.
	 */
	public abstract int getPoints();
	
	/**
	 * Get the type of the object family
	 * 
	 * @return Type of object
	 */
	public abstract int getType();

	/**
	 * Mark the target as currently selected.
	 * 
	 * @param b True if this target is selected.
	 */
	public void setCurrentTarget(boolean b) {
	
		currentTarget = b;
	}
	
	/**
	 * Returns true if this target is currently selected.
	 * 
	 * @return True if selected.
	 */
	public boolean isCurrentTarget() {
	
		return currentTarget;
	}
	
	/**
	 * Mark the target as possible colliding with the ship.
	 * 
	 * @param b True if this target may collide.
	 */
	public void setCritical(boolean b) {
	
		critical = b;
	}
	
	/**
	 * see GameObject#merge(GameObject)
	 * 
	 * @param t The predecessor to merge data from.
	 * @param history A list of historic frames.
	 */
	public void merge(Target t, LinkedList<FrameInfo> history) {

		super.merge(t, history);
		
		// merge shot-marker
		if (t.isUnderFire()) 
			setUnderFire(t.getMarkedAsShotAtFrame());
		// merge shooting solution
		if (t.hasShootingSolution())
			setShootingSolution(t.getShootingSolution());
	}
	
	/**
	 * Returns true if this target may collide with the ship.
	 * 
	 * @return True if this target may collide.
	 */
	public boolean isCritical() {
		
		return critical;
	}
	
	/**
	 * @see GameObject#draw(Graphics2D)
	 */
	@Override
	public void draw(Graphics2D g) {
		
		super.draw(g);
		
		// draw target uid
		g.setColor(Color.gray);
		AffineTransform trafo = new AffineTransform();
		trafo.concatenate(AffineTransform.getScaleInstance(1, -1));
		g.setFont(new Font("Dialog", Font.PLAIN, 120).deriveFont(trafo));
		g.drawString(Integer.toString(getUid()), x + 120, y - 80);
	    
		// TODO so ein blödes rumgefrickel an symptomen - das darf doch niemals
		// vorkommen, dass es solutions ohne hitpoint gibt!!!!!!
	    if (shootingSolution != null && shootingSolution.getHitPoint() != null) {
	    	// draw hitpoint 
	    	g.setColor(Color.red);
	    	g.fillOval(
	    			shootingSolution.getHitPoint().x - 8, 
	    			shootingSolution.getHitPoint().y - 8, 
	    			16, 
	    			16);
	    }
	}
	
	/**
	 * Get the appropriate color for drawing
	 * 
	 * @return Color for drawing
	 */
	protected Color getColor() {

		if (underFire)
			return Color.magenta;
		else if (critical)
			return Color.cyan;
		else if (currentTarget) 
			return Color.green;
		else
			return normalColor;
	}
	
	/**
	 * Mark this target as shot.
	 * 
	 * @param underFire True if this target has been shot.
	 * @param frameNo The frame number when the shot was fired.
	 */
	public void setUnderFire(int frameNo) {
		
		this.underFire = true;
		shotTime = frameNo;
	}

	/**
	 * Removes the target-is-under-fire-flag.
	 */
	public void removeMarkedAsShot() {

		underFire = false;
		shotTime = -1;
	}
	
	/**
	 * Returns true if this target is marked as shot.
	 * 
	 * @return True if this target has been shot.
	 */
	public boolean isUnderFire() {
		
		return underFire;
	}
	
	/**
	 * Returns the frame number when this target has been marked as shot.
	 * 
	 * @return The frame number.
	 */
	public int getMarkedAsShotAtFrame() {
		
		return shotTime;
	}
	
//	/**
//	 * Returns true if this object equals the given object by size, type and
//	 * velocity.
//	 * 
//	 * @param t The target to compare.
//	 * @return True if both are equal.
//	 */
//	public boolean equals(Object t)
//    {
//		
//		if (this == t) {
//			return true;
//		}
//      
//		if (t == null || getClass() != t.getClass()) {
//			return false;
//		}
//      
//		Target target = (Target) t;
//		
//		if ((int)getVelocity() != (int)target.getVelocity()) {
//			return false;
//		}
//      
//		if (Math.abs(getVelocityAngle() - target.getVelocityAngle()) > Math.PI/60) {
//			return false;
//		}
//
//		return true;
//    }

	/**
	 * @return the shootingSolution
	 */
	public ShootingSolution getShootingSolution() {
		
		return shootingSolution;
	}

	/**
	 * @param shootingSolution the shootingSolution to set
	 */
	public void setShootingSolution(ShootingSolution shootingSolution) {
		
		this.shootingSolution = shootingSolution;
	}
	
	/**
	 * Checks whether a valid shooting solution exists or not
	 * 
	 * @return TRUE if a valid shooting solution exists
	 */
	public boolean hasShootingSolution() {
		
		return (shootingSolution != null);
	}
}
